import os
import time
import pdb
import re

import gradio as gr
import numpy as np
import sys
import subprocess

from huggingface_hub import snapshot_download
import requests

import argparse
import os
from omegaconf import OmegaConf
import numpy as np
import cv2
import torch
import glob
import pickle
from tqdm import tqdm
import copy
from argparse import Namespace
import shutil
import gdown
import imageio
import ffmpeg
from moviepy.editor import *
from transformers import WhisperModel

ProjectDir = os.path.abspath(os.path.dirname(__file__))
CheckpointsDir = os.path.join(ProjectDir, "models")

@torch.no_grad()
def debug_inpainting(video_path, bbox_shift, extra_margin=10, parsing_mode="jaw", 
                    left_cheek_width=90, right_cheek_width=90):
    """Debug inpainting parameters, only process the first frame"""
    # Set default parameters
    args_dict = {
        "result_dir": './results/debug', 
        "fps": 25, 
        "batch_size": 1, 
        "output_vid_name": '', 
        "use_saved_coord": False,
        "audio_padding_length_left": 2,
        "audio_padding_length_right": 2,
        "version": "v15",
        "extra_margin": extra_margin,
        "parsing_mode": parsing_mode,
        "left_cheek_width": left_cheek_width,
        "right_cheek_width": right_cheek_width
    }
    args = Namespace(**args_dict)

    # Create debug directory
    os.makedirs(args.result_dir, exist_ok=True)
    
    # Read first frame
    if get_file_type(video_path) == "video":
        reader = imageio.get_reader(video_path)
        first_frame = reader.get_data(0)
        reader.close()
    else:
        first_frame = cv2.imread(video_path)
        first_frame = cv2.cvtColor(first_frame, cv2.COLOR_BGR2RGB)
    
    # Save first frame
    debug_frame_path = os.path.join(args.result_dir, "debug_frame.png")
    cv2.imwrite(debug_frame_path, cv2.cvtColor(first_frame, cv2.COLOR_RGB2BGR))
    
    # Get face coordinates
    coord_list, frame_list = get_landmark_and_bbox([debug_frame_path], bbox_shift)
    bbox = coord_list[0]
    frame = frame_list[0]
    
    if bbox == coord_placeholder:
        return None, "No face detected, please adjust bbox_shift parameter"
    
    # Initialize face parser
    fp = FaceParsing(
        left_cheek_width=args.left_cheek_width,
        right_cheek_width=args.right_cheek_width
    )
    
    # Process first frame
    x1, y1, x2, y2 = bbox
    y2 = y2 + args.extra_margin
    y2 = min(y2, frame.shape[0])
    crop_frame = frame[y1:y2, x1:x2]
    crop_frame = cv2.resize(crop_frame,(256,256),interpolation = cv2.INTER_LANCZOS4)
    
    # Generate random audio features
    random_audio = torch.randn(1, 50, 384, device=device, dtype=weight_dtype)
    audio_feature = pe(random_audio)
    
    # Get latents
    latents = vae.get_latents_for_unet(crop_frame)
    latents = latents.to(dtype=weight_dtype)
    
    # Generate prediction results
    pred_latents = unet.model(latents, timesteps, encoder_hidden_states=audio_feature).sample
    recon = vae.decode_latents(pred_latents)
    
    # Inpaint back to original image
    res_frame = recon[0]
    res_frame = cv2.resize(res_frame.astype(np.uint8),(x2-x1,y2-y1))
    combine_frame = get_image(frame, res_frame, [x1, y1, x2, y2], mode=args.parsing_mode, fp=fp)
    
    # Save results (no need to convert color space again since get_image already returns RGB format)
    debug_result_path = os.path.join(args.result_dir, "debug_result.png")
    cv2.imwrite(debug_result_path, combine_frame)
    
    # Create information text
    info_text = f"Parameter information:\n" + \
                f"bbox_shift: {bbox_shift}\n" + \
                f"extra_margin: {extra_margin}\n" + \
                f"parsing_mode: {parsing_mode}\n" + \
                f"left_cheek_width: {left_cheek_width}\n" + \
                f"right_cheek_width: {right_cheek_width}\n" + \
                f"Detected face coordinates: [{x1}, {y1}, {x2}, {y2}]"
    
    return cv2.cvtColor(combine_frame, cv2.COLOR_RGB2BGR), info_text

def print_directory_contents(path):
    for child in os.listdir(path):
        child_path = os.path.join(path, child)
        if os.path.isdir(child_path):
            print(child_path)

def download_model():
    # 检查必需的模型文件是否存在
    required_models = {
        "MuseTalk": f"{CheckpointsDir}/musetalkV15/unet.pth",
        "MuseTalk": f"{CheckpointsDir}/musetalkV15/musetalk.json",
        "SD VAE": f"{CheckpointsDir}/sd-vae/config.json",
        "Whisper": f"{CheckpointsDir}/whisper/config.json",
        "DWPose": f"{CheckpointsDir}/dwpose/dw-ll_ucoco_384.pth",
        "SyncNet": f"{CheckpointsDir}/syncnet/latentsync_syncnet.pt",
        "Face Parse": f"{CheckpointsDir}/face-parse-bisent/79999_iter.pth",
        "ResNet": f"{CheckpointsDir}/face-parse-bisent/resnet18-5c106cde.pth"
    }
    
    missing_models = []
    for model_name, model_path in required_models.items():
        if not os.path.exists(model_path):
            missing_models.append(model_name)
    
    if missing_models:
        # 全用英文
        print("The following required model files are missing:")
        for model in missing_models:
            print(f"- {model}")
        print("\nPlease run the download script to download the missing models:")
        if sys.platform == "win32":
            print("Windows: Run download_weights.bat")
        else:
            print("Linux/Mac: Run ./download_weights.sh")
        sys.exit(1)
    else:
        print("All required model files exist.")




download_model()  # for huggingface deployment.

from musetalk.utils.blending import get_image
from musetalk.utils.face_parsing import FaceParsing
from musetalk.utils.audio_processor import AudioProcessor
from musetalk.utils.utils import get_file_type, get_video_fps, datagen, load_all_model
from musetalk.utils.preprocessing import get_landmark_and_bbox, read_imgs, coord_placeholder, get_bbox_range


def fast_check_ffmpeg():
    try:
        subprocess.run(["ffmpeg", "-version"], capture_output=True, check=True)
        return True
    except:
        return False


@torch.no_grad()
def inference(audio_path, video_path, bbox_shift, extra_margin=10, parsing_mode="jaw", 
              left_cheek_width=90, right_cheek_width=90, progress=gr.Progress(track_tqdm=True)):
    # Set default parameters, aligned with inference.py
    args_dict = {
        "result_dir": './results/output', 
        "fps": 25, 
        "batch_size": 8, 
        "output_vid_name": '', 
        "use_saved_coord": False,
        "audio_padding_length_left": 2,
        "audio_padding_length_right": 2,
        "version": "v15",  # Fixed use v15 version
        "extra_margin": extra_margin,
        "parsing_mode": parsing_mode,
        "left_cheek_width": left_cheek_width,
        "right_cheek_width": right_cheek_width
    }
    args = Namespace(**args_dict)

    # Check ffmpeg
    if not fast_check_ffmpeg():
        print("Warning: Unable to find ffmpeg, please ensure ffmpeg is properly installed")

    input_basename = os.path.basename(video_path).split('.')[0]
    audio_basename = os.path.basename(audio_path).split('.')[0]
    output_basename = f"{input_basename}_{audio_basename}"
    
    # Create temporary directory
    temp_dir = os.path.join(args.result_dir, f"{args.version}")
    os.makedirs(temp_dir, exist_ok=True)
    
    # Set result save path
    result_img_save_path = os.path.join(temp_dir, output_basename)
    crop_coord_save_path = os.path.join(args.result_dir, "../", input_basename+".pkl")
    os.makedirs(result_img_save_path, exist_ok=True)

    if args.output_vid_name == "":
        output_vid_name = os.path.join(temp_dir, output_basename+".mp4")
    else:
        output_vid_name = os.path.join(temp_dir, args.output_vid_name)
        
    ############################################## extract frames from source video ##############################################
    if get_file_type(video_path) == "video":
        save_dir_full = os.path.join(temp_dir, input_basename)
        os.makedirs(save_dir_full, exist_ok=True)
        # Read video
        reader = imageio.get_reader(video_path)

        # Save images
        for i, im in enumerate(reader):
            imageio.imwrite(f"{save_dir_full}/{i:08d}.png", im)
        input_img_list = sorted(glob.glob(os.path.join(save_dir_full, '*.[jpJP][pnPN]*[gG]')))
        fps = get_video_fps(video_path)
    else: # input img folder
        input_img_list = glob.glob(os.path.join(video_path, '*.[jpJP][pnPN]*[gG]'))
        input_img_list = sorted(input_img_list, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        fps = args.fps
        
    ############################################## extract audio feature ##############################################
    # Extract audio features
    whisper_input_features, librosa_length = audio_processor.get_audio_feature(audio_path)
    whisper_chunks = audio_processor.get_whisper_chunk(
        whisper_input_features, 
        device, 
        weight_dtype, 
        whisper, 
        librosa_length,
        fps=fps,
        audio_padding_length_left=args.audio_padding_length_left,
        audio_padding_length_right=args.audio_padding_length_right,
    )
        
    ############################################## preprocess input image  ##############################################
    if os.path.exists(crop_coord_save_path) and args.use_saved_coord:
        print("using extracted coordinates")
        with open(crop_coord_save_path,'rb') as f:
            coord_list = pickle.load(f)
        frame_list = read_imgs(input_img_list)
    else:
        print("extracting landmarks...time consuming")
        coord_list, frame_list = get_landmark_and_bbox(input_img_list, bbox_shift)
        with open(crop_coord_save_path, 'wb') as f:
            pickle.dump(coord_list, f)
    bbox_shift_text = get_bbox_range(input_img_list, bbox_shift)
    
    # Initialize face parser
    fp = FaceParsing(
        left_cheek_width=args.left_cheek_width,
        right_cheek_width=args.right_cheek_width
    )
    
    i = 0
    input_latent_list = []
    for bbox, frame in zip(coord_list, frame_list):
        if bbox == coord_placeholder:
            continue
        x1, y1, x2, y2 = bbox
        y2 = y2 + args.extra_margin
        y2 = min(y2, frame.shape[0])
        crop_frame = frame[y1:y2, x1:x2]
        crop_frame = cv2.resize(crop_frame,(256,256),interpolation = cv2.INTER_LANCZOS4)
        latents = vae.get_latents_for_unet(crop_frame)
        input_latent_list.append(latents)

    # to smooth the first and the last frame
    frame_list_cycle = frame_list + frame_list[::-1]
    coord_list_cycle = coord_list + coord_list[::-1]
    input_latent_list_cycle = input_latent_list + input_latent_list[::-1]
    
    ############################################## inference batch by batch ##############################################
    print("start inference")
    video_num = len(whisper_chunks)
    batch_size = args.batch_size
    gen = datagen(
        whisper_chunks=whisper_chunks,
        vae_encode_latents=input_latent_list_cycle,
        batch_size=batch_size,
        delay_frame=0,
        device=device,
    )
    res_frame_list = []
    for i, (whisper_batch,latent_batch) in enumerate(tqdm(gen,total=int(np.ceil(float(video_num)/batch_size)))):
        audio_feature_batch = pe(whisper_batch)
        # Ensure latent_batch is consistent with model weight type
        latent_batch = latent_batch.to(dtype=weight_dtype)
        
        pred_latents = unet.model(latent_batch, timesteps, encoder_hidden_states=audio_feature_batch).sample
        recon = vae.decode_latents(pred_latents)
        for res_frame in recon:
            res_frame_list.append(res_frame)
            
    ############################################## pad to full image ##############################################
    print("pad talking image to original video")
    for i, res_frame in enumerate(tqdm(res_frame_list)):
        bbox = coord_list_cycle[i%(len(coord_list_cycle))]
        ori_frame = copy.deepcopy(frame_list_cycle[i%(len(frame_list_cycle))])
        x1, y1, x2, y2 = bbox
        y2 = y2 + args.extra_margin
        y2 = min(y2, frame.shape[0])
        try:
            res_frame = cv2.resize(res_frame.astype(np.uint8),(x2-x1,y2-y1))
        except:
            continue
        
        # Use v15 version blending
        combine_frame = get_image(ori_frame, res_frame, [x1, y1, x2, y2], mode=args.parsing_mode, fp=fp)
            
        cv2.imwrite(f"{result_img_save_path}/{str(i).zfill(8)}.png",combine_frame)
        
    # Frame rate
    fps = 25
    # Output video path
    output_video = 'temp.mp4'

    # Read images
    def is_valid_image(file):
        pattern = re.compile(r'\d{8}\.png')
        return pattern.match(file)

    images = []
    files = [file for file in os.listdir(result_img_save_path) if is_valid_image(file)]
    files.sort(key=lambda x: int(x.split('.')[0]))

    for file in files:
        filename = os.path.join(result_img_save_path, file)
        images.append(imageio.imread(filename))
        

    # Save video
    imageio.mimwrite(output_video, images, 'FFMPEG', fps=fps, codec='libx264', pixelformat='yuv420p')

    input_video = './temp.mp4'
    # Check if the input_video and audio_path exist
    if not os.path.exists(input_video):
        raise FileNotFoundError(f"Input video file not found: {input_video}")
    if not os.path.exists(audio_path):
        raise FileNotFoundError(f"Audio file not found: {audio_path}")
    
    # Read video
    reader = imageio.get_reader(input_video)
    fps = reader.get_meta_data()['fps']  # Get original video frame rate
    reader.close() # Otherwise, error on win11: PermissionError: [WinError 32] Another program is using this file, process cannot access. : 'temp.mp4'
    # Store frames in list
    frames = images
    
    print(len(frames))

    # Load the video
    video_clip = VideoFileClip(input_video)

    # Load the audio
    audio_clip = AudioFileClip(audio_path)

    # Set the audio to the video
    video_clip = video_clip.set_audio(audio_clip)

    # Write the output video
    video_clip.write_videofile(output_vid_name, codec='libx264', audio_codec='aac',fps=25)

    os.remove("temp.mp4")
    #shutil.rmtree(result_img_save_path)
    print(f"result is save to {output_vid_name}")
    return output_vid_name,bbox_shift_text



# load model weights
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
vae, unet, pe = load_all_model(
    unet_model_path="./models/musetalkV15/unet.pth", 
    vae_type="sd-vae",
    unet_config="./models/musetalkV15/musetalk.json",
    device=device
)

# Parse command line arguments
parser = argparse.ArgumentParser()
parser.add_argument("--ffmpeg_path", type=str, default=r"ffmpeg-master-latest-win64-gpl-shared\bin", help="Path to ffmpeg executable")
parser.add_argument("--ip", type=str, default="127.0.0.1", help="IP address to bind to")
parser.add_argument("--port", type=int, default=7860, help="Port to bind to")
parser.add_argument("--share", action="store_true", help="Create a public link")
parser.add_argument("--use_float16", action="store_true", help="Use float16 for faster inference")
args = parser.parse_args()

# Set data type
if args.use_float16:
    # Convert models to half precision for better performance
    pe = pe.half()
    vae.vae = vae.vae.half()
    unet.model = unet.model.half()
    weight_dtype = torch.float16
else:
    weight_dtype = torch.float32

# Move models to specified device
pe = pe.to(device)
vae.vae = vae.vae.to(device)
unet.model = unet.model.to(device)

timesteps = torch.tensor([0], device=device)

# Initialize audio processor and Whisper model
audio_processor = AudioProcessor(feature_extractor_path="./models/whisper")
whisper = WhisperModel.from_pretrained("./models/whisper")
whisper = whisper.to(device=device, dtype=weight_dtype).eval()
whisper.requires_grad_(False)


def check_video(video):
    if not isinstance(video, str):
        return video # in case of none type
    # Define the output video file name
    dir_path, file_name = os.path.split(video)
    if file_name.startswith("outputxxx_"):
        return video
    # Add the output prefix to the file name
    output_file_name = "outputxxx_" + file_name

    os.makedirs('./results',exist_ok=True)
    os.makedirs('./results/output',exist_ok=True)
    os.makedirs('./results/input',exist_ok=True)

    # Combine the directory path and the new file name
    output_video = os.path.join('./results/input', output_file_name)


    # read video
    reader = imageio.get_reader(video)
    fps = reader.get_meta_data()['fps']  # get fps from original video

    # conver fps to 25
    frames = [im for im in reader]
    target_fps = 25
    
    L = len(frames)
    L_target = int(L / fps * target_fps)
    original_t = [x / fps for x in range(1, L+1)]
    t_idx = 0
    target_frames = []
    for target_t in range(1, L_target+1):
        while target_t / target_fps > original_t[t_idx]:
            t_idx += 1      # find the first t_idx so that target_t / target_fps <= original_t[t_idx]
            if t_idx >= L:
                break
        target_frames.append(frames[t_idx])

    # save video
    imageio.mimwrite(output_video, target_frames, 'FFMPEG', fps=25, codec='libx264', quality=9, pixelformat='yuv420p')
    return output_video




css = """#input_img {max-width: 1024px !important} #output_vid {max-width: 1024px; max-height: 576px}"""

with gr.Blocks(css=css) as demo:
    gr.Markdown(
        """<div align='center'> <h1>MuseTalk: Real-Time High-Fidelity Video Dubbing via Spatio-Temporal Sampling</h1> \
                    <h2 style='font-weight: 450; font-size: 1rem; margin: 0rem'>\
                    </br>\
                    Yue Zhang <sup>*</sup>,\
                    Zhizhou Zhong <sup>*</sup>,\
                    Minhao Liu<sup>*</sup>,\
                    Zhaokang Chen,\
                    Bin Wu<sup>†</sup>,\
                    Yubin Zeng,\
                    Chao Zhang,\
                    Yingjie He,\
                    Junxin Huang,\
                    Wenjiang Zhou <br>\
                    (<sup>*</sup>Equal Contribution, <sup>†</sup>Corresponding Author, benbinwu@tencent.com)\
                    Lyra Lab, Tencent Music Entertainment\
                </h2> \
                <a style='font-size:18px;color: #000000' href='https://github.com/TMElyralab/MuseTalk'>[Github Repo]</a>\
                <a style='font-size:18px;color: #000000' href='https://github.com/TMElyralab/MuseTalk'>[Huggingface]</a>\
                <a style='font-size:18px;color: #000000' href='https://arxiv.org/abs/2410.10122'> [Technical report] </a>"""
    )

    with gr.Row():
        with gr.Column():
            audio = gr.Audio(label="Drving Audio",type="filepath")
            video = gr.Video(label="Reference Video",sources=['upload'])
            bbox_shift = gr.Number(label="BBox_shift value, px", value=0)
            extra_margin = gr.Slider(label="Extra Margin", minimum=0, maximum=40, value=10, step=1)
            parsing_mode = gr.Radio(label="Parsing Mode", choices=["jaw", "raw"], value="jaw")
            left_cheek_width = gr.Slider(label="Left Cheek Width", minimum=20, maximum=160, value=90, step=5)
            right_cheek_width = gr.Slider(label="Right Cheek Width", minimum=20, maximum=160, value=90, step=5)
            bbox_shift_scale = gr.Textbox(label="'left_cheek_width' and 'right_cheek_width' parameters determine the range of left and right cheeks editing when parsing model is 'jaw'. The 'extra_margin' parameter determines the movement range of the jaw. Users can freely adjust these three parameters to obtain better inpainting results.")

            with gr.Row():
                debug_btn = gr.Button("1. Test Inpainting ")
                btn = gr.Button("2. Generate")
        with gr.Column():
            debug_image = gr.Image(label="Test Inpainting Result (First Frame)")
            debug_info = gr.Textbox(label="Parameter Information", lines=5)
            out1 = gr.Video()
    
    video.change(
        fn=check_video, inputs=[video], outputs=[video]
    )
    btn.click(
        fn=inference,
        inputs=[
            audio,
            video,
            bbox_shift,
            extra_margin,
            parsing_mode,
            left_cheek_width,
            right_cheek_width
        ],
        outputs=[out1,bbox_shift_scale]
    )
    debug_btn.click(
        fn=debug_inpainting,
        inputs=[
            video,
            bbox_shift,
            extra_margin,
            parsing_mode,
            left_cheek_width,
            right_cheek_width
        ],
        outputs=[debug_image, debug_info]
    )

# Check ffmpeg and add to PATH
if not fast_check_ffmpeg():
    print(f"Adding ffmpeg to PATH: {args.ffmpeg_path}")
    # According to operating system, choose path separator
    path_separator = ';' if sys.platform == 'win32' else ':'
    os.environ["PATH"] = f"{args.ffmpeg_path}{path_separator}{os.environ['PATH']}"
    if not fast_check_ffmpeg():
        print("Warning: Unable to find ffmpeg, please ensure ffmpeg is properly installed")

# Solve asynchronous IO issues on Windows
if sys.platform == 'win32':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

# Start Gradio application
demo.queue().launch(
    share=args.share, 
    debug=True, 
    server_name=args.ip, 
    server_port=args.port
)



import logging
from time import perf_counter
from typing import Generator, Optional

import torch

try:
    from transformers import (
        Qwen3OmniMoeForConditionalGeneration,
        Qwen3OmniMoeConfig,
        Qwen3OmniMoeCode2Wav,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Qwen3Omni, you need to `pip install git+https://github.com/huggingface/transformers`"
    )
    raise Exception(f"Missing module: {e}")


class Qwen3OmniMoeCode2WavStreaming(Qwen3OmniMoeCode2Wav):
    def chunked_decode_stream(self, codes, chunk_size=50, left_context_size=25):
        start_index = 0
        while start_index < codes.shape[-1]:
            end_index = min(start_index + chunk_size, codes.shape[-1])
            context_size = left_context_size if start_index - left_context_size > 0 else start_index
            codes_chunk = codes[..., start_index - context_size : end_index]
            wav_chunk = self(codes_chunk)
            yield wav_chunk[..., context_size * self.total_upsample :]
            start_index = end_index


class Qwen3OmniMoeForConditionalGenerationStreaming(Qwen3OmniMoeForConditionalGeneration):
    def __init__(self, config: Qwen3OmniMoeConfig):
        super().__init__(config)
        if hasattr(self, "code2wav"):
            self.code2wav = Qwen3OmniMoeCode2WavStreaming(config.code2wav_config)

    @torch.no_grad()
    def thinker_generate_chunk(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_max_tokens_per_step: int = 10,  # Controls how many tokens to generate *per step*
        thinker_max_new_tokens: int = 1024,
        thinker_top_k: int = 20,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_eos_token_ids: list = [151643, 151645],  # Define EOS tokens
        thinker_repetition_penalty: float = 1.05,
        thinker_output_hidden_states: bool = False,
        thinker_stop_strings_per_step: list = [],
        tokenizer=None,
        **kwargs,
    ):
        """
        {
            "thinker_generate_ids": # [1,seq_len],
            "thinker_generate_hidden_states":  # tuple(tuple(tensor[1, 1, 2048])...)
            "full_generated_ids": # tuple(tuple(tensor[1, 1, 2048])...)
        }
        """
        input_ids = inputs.get("input_ids")
        attention_mask = inputs.get("attention_mask", None)

        if thinker_max_tokens_per_step > thinker_max_new_tokens:
            thinker_max_tokens_per_step = thinker_max_new_tokens

        # Keep track of the full generated sequence full_generated_ids = input_ids.clone()
        # Ensure full_attention_mask is correctly initialized and expanded
        full_attention_mask = (
            attention_mask.clone()
            if attention_mask is not None
            else torch.ones_like(input_ids, device=input_ids.device)
        )
        full_generated_ids = input_ids.clone()

        # KV cache
        # past_key_values = None

        # Inputs for the current step
        current_input_ids = full_generated_ids
        # The attention mask passed to generate should cover the sequence length for the current step
        current_attention_mask = full_attention_mask

        total_new_tokens_generated = 0
        hidden_states = None
        hidden_states_len = 0

        times = []
        while total_new_tokens_generated < thinker_max_new_tokens:
            # Prepare inputs for generate call
            # logging.info(current_input_ids, current_attention_mask.shape)
            model_inputs = {
                "input_ids": current_input_ids,
                "attention_mask": current_attention_mask,
                # "past_key_values": past_key_values,
                "use_cache": True,
                "use_audio_in_video": use_audio_in_video,
                "do_sample": True if thinker_temperature > 0 else False,
                "top_k": thinker_top_k,
                "top_p": thinker_top_p,
                "temperature": thinker_temperature,
                "repetition_penalty": thinker_repetition_penalty,
                "min_new_tokens": 1,  # Ensure at least one token is generated if possible
                "max_new_tokens": thinker_max_tokens_per_step,  # Generate in smaller steps
                # output_hidden_states/scores can consume memory,
                # enable if needed downstream(talker)
                "output_hidden_states": thinker_output_hidden_states,
                "return_dict_in_generate": True,
                # "output_scores": True,
                "eos_token_id": thinker_eos_token_ids,
                "pad_token_id": kwargs.get("thinker_pad_token_id", 151643),
            }
            model_inputs = {**inputs, **model_inputs}
            model_inputs["input_ids"] = current_input_ids
            model_inputs["attention_mask"] = current_attention_mask
            if len(thinker_stop_strings_per_step) > 0:
                model_inputs["stop_strings"] = thinker_stop_strings_per_step
                model_inputs["tokenizer"] = tokenizer

            start_time = perf_counter()
            outputs = self.thinker.generate(**model_inputs)
            times.append(perf_counter() - start_time)

            # Extract newly generated token IDs *for this step*
            # `outputs.sequences` contains the input_ids for this step + new tokens generated in this step
            step_new_ids = outputs.sequences[:, current_input_ids.shape[1] :]
            num_step_new_tokens = step_new_ids.shape[1]
            total_new_tokens_generated += num_step_new_tokens

            if num_step_new_tokens == 0:  # Handle case where generate stops early
                logging.warning("Warning: generate produced 0 new tokens in this step.")
                break

            if thinker_output_hidden_states is True:
                hidden_states = outputs.hidden_states
                hidden_states_len = (
                    hidden_states_len if hidden_states_len > 0 else hidden_states[0][0].shape[1]
                )
                # new generate thinker_token_embeds
                thinker_new_token_embeds = hidden_states[0][0][:, :hidden_states_len, :]
                hidden_states = (
                    (thinker_new_token_embeds,) + hidden_states[0][1:],
                ) + hidden_states[1:]
                # new generate thinker_hidden_states
                thinker_new_hidden_states = hidden_states[0][-1][:, :hidden_states_len, :]
                hidden_states = (
                    hidden_states[0][:-1] + (thinker_new_hidden_states,),
                ) + hidden_states[1:]

            # Update the full sequence
            full_generated_ids = torch.cat([full_generated_ids, step_new_ids], dim=1)

            yield {
                "thinker_generate_ids": step_new_ids,
                "thinker_generate_hidden_states": hidden_states,
                "full_generated_ids": full_generated_ids,
            }

            # Prepare for the next iteration:
            # Input is only the last generated token
            # NOTE: need use past_key_values to keep the context by manually,
            # current_input_ids = step_new_ids[:, -1:]
            # so we can't use the last generated token, use cache instead
            # input ids need to be the full sequence for next generation
            current_input_ids = full_generated_ids

            # Update past_key_values
            # past_key_values = outputs.past_key_values

            # Update attention mask by appending 1s for the new tokens
            full_attention_mask = torch.cat(
                [full_attention_mask, torch.ones_like(step_new_ids)], dim=1
            )
            current_attention_mask = full_attention_mask

            # Check if EOS token was generated in this step
            if step_new_ids[0, -1].item() in thinker_eos_token_ids:
                logging.info("EOS token generated.")
                break

            # Check if max_new_tokens limit is reached (after processing the step)
            if total_new_tokens_generated >= thinker_max_new_tokens:
                logging.info("Max new tokens limit reached.")
                break

        logging.info(
            f"Total new tokens generated: {total_new_tokens_generated} | thinker_max_tokens_per_step: {thinker_max_tokens_per_step} | first chunk generated cost: {times[0]} s | total cost: {sum(times)} s"
        )

    @torch.no_grad()
    def talker_generate(
        self,
        hidden_states: tuple,
        sequences: torch.Tensor,
        input_ids: torch.Tensor,
        speaker_id: int,
        talker_kwargs: dict,
        token2wav_kwargs: dict,
    ) -> Generator[torch.Tensor, None, None]:
        # 2. Prepare talker input
        thinker_embed = torch.cat([h[0] for h in hidden_states], dim=1).to(
            self.talker.device
        )  # [1 t d]
        thinker_hidden = torch.cat(
            [
                h[self.config.talker_config.accept_hidden_layer] for h in hidden_states
            ],  # accept_hidden_layer:24
            dim=1,
        ).to(self.talker.device)  # [1 t d]
        im_start_indexes = torch.cat(
            (
                torch.nonzero(input_ids[0] == self.config.im_start_token_id).squeeze(),
                torch.tensor(
                    [sequences.shape[-1]],
                    device=input_ids.device,
                    dtype=input_ids.dtype,
                ),
            ),
            dim=-1,
        ).to(
            self.talker.device
        )  # Shape [n_starts + 1]; Take batch 0 since batched inference is not supported here.
        multimodal_mask = (
            (sequences == self.config.thinker_config.audio_token_id) |
            (sequences == self.config.thinker_config.image_token_id) |
            (sequences == self.config.thinker_config.video_token_id)
        ).to(self.talker.device)  # [1 t] # fmt: skip

        talker_special_tokens = torch.tensor(
            [
                [
                    self.config.tts_bos_token_id,
                    self.config.tts_eos_token_id,
                    self.config.tts_pad_token_id,
                ]
            ],
            device=self.thinker.device,
            dtype=input_ids.dtype,
        )
        tts_bos_embed, tts_eos_embed, tts_pad_embed = (
            self.talker.text_projection(self.thinker.get_input_embeddings()(talker_special_tokens))
            .to(self.talker.device)
            .chunk(3, dim=1)
        )  # 3 * [1 1 d]

        talker_input_embeds = []  # [1 t d]
        talker_input_ids = []
        # For every chatml parts
        for i in range(len(im_start_indexes) - 1):
            im_start_index = im_start_indexes[i]
            segment_end_index = im_start_indexes[i + 1]
            role_token = input_ids[0][im_start_index + 1]
            # Talker should ignore thinker system prompt
            if role_token == self.config.system_token_id:
                continue
            # Talker takes word embeddings for tokens and hidden state from `accept_hidden_layer` for multimodal inputs
            elif role_token == self.config.user_token_id:
                talker_user_part = self._get_talker_user_parts(
                    im_start_index,
                    segment_end_index,
                    multimodal_mask,
                    thinker_hidden,
                    thinker_embed,
                )
                talker_input_embeds.append(talker_user_part)
                talker_input_ids.append(sequences[:, im_start_index:segment_end_index])
            # Take assistant output (for now)
            elif role_token == self.config.assistant_token_id and i == len(im_start_indexes) - 2:
                talker_assistant_embeds, talker_assistant_ids, trailing_text_hidden = (
                    self._get_talker_assistant_parts(
                        im_start_index,
                        segment_end_index,
                        speaker_id,
                        thinker_embed,
                        tts_pad_embed,
                        tts_bos_embed,
                        tts_eos_embed,
                    )
                )
                talker_input_embeds.append(talker_assistant_embeds)
                talker_input_ids.append(talker_assistant_ids)
            # History assistant output (ignore for now)
            elif role_token == self.config.assistant_token_id and i != len(im_start_indexes) - 2:
                continue
            else:
                raise AssertionError("Expect role id after <|im_start|> (assistant, user, system)")
        talker_input_embed = torch.cat(
            [embed.to(self.talker.device) for embed in talker_input_embeds], dim=1
        )
        talker_input_id = torch.cat(
            [embed.to(self.talker.device) for embed in talker_input_ids], dim=1
        )

        talker_result = self.talker.generate(
            attention_mask=torch.ones(talker_input_id.shape, device=talker_input_id.device),
            pad_token_id=2150,
            inputs_embeds=talker_input_embed,
            trailing_text_hidden=trailing_text_hidden,
            tts_pad_embed=tts_pad_embed,
            talker_input_ids=talker_input_id,  # Not use input_ids to prevent repetation penalty out of bound
            **talker_kwargs,
        )

        talker_codes = (
            torch.stack(
                [hid[-1] for hid in talker_result.hidden_states if hid[-1] is not None], dim=1
            )
            .transpose(1, 2)
            .to(self.code2wav.device)
        )

        chunk_size = token2wav_kwargs.get("chunk_size", 50)
        left_context_size = token2wav_kwargs.get("left_context_size", 25)
        return self.code2wav.chunked_decode_stream(
            talker_codes, chunk_size=chunk_size, left_context_size=left_context_size
        )

    @torch.no_grad()
    def generate_stream(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
        thinker_max_new_tokens: int = 1024,
        thinker_top_k: int = 20,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_repetition_penalty: float = 1.05,
        thinker_eos_token_ids: list[int] = [151643, 151645],
        thinker_stop_strings_per_step: list[str] = [],
        tokenizer=None,
        return_audio: bool = True,
        speaker: str = "Chelsie",
        talker_top_k: int = 10,
        talker_top_p: float = 0.9,
        talker_temperature: float = 0.95,
        talker_repetition_penalty: float = 1.1,
        talker_max_new_tokens: int = 8192,
        token2wav_kwargs: dict = {"chunk_size": 50, "left_context_size": 25},
        skip_token_ids: Optional[list] = [],
        **kwargs,
    ) -> Generator[dict, None, None]:
        """
        - return Generator[dict, None, None]
        {
            "thinker_ids": torch.Tensor, # (1,T)
            "talker_wav": torch.Tensor, # (1,1,T)
        }
        """
        input_ids = inputs.get("input_ids")
        thinker_chunk_stream = self.thinker_generate_chunk(
            inputs,
            use_audio_in_video=use_audio_in_video,
            thinker_max_tokens_per_step=thinker_max_tokens_per_step,
            thinker_max_new_tokens=thinker_max_new_tokens,
            thinker_top_k=thinker_top_k,
            thinker_top_p=thinker_top_p,
            thinker_temperature=thinker_temperature,
            thinker_eos_token_ids=thinker_eos_token_ids,
            thinker_repetition_penalty=thinker_repetition_penalty,
            thinker_output_hidden_states=return_audio,
            thinker_stop_strings_per_step=thinker_stop_strings_per_step,
            tokenizer=tokenizer,
            **kwargs,
        )
        if not return_audio:
            for thinker_chunk in thinker_chunk_stream:
                yield {"thinker_ids": thinker_chunk["thinker_generate_ids"]}
        else:
            speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
            if speaker_id is None:
                raise NotImplementedError(f"Speaker {speaker} not implemented")
            if input_ids.shape[0] != 1:
                raise NotImplementedError(
                    "Qwen3-Omni currently does not support batched inference with audio output"
                )
            talker_supppressed_tokens = [
                i
                for i in range(
                    self.config.talker_config.text_config.vocab_size - 1024,
                    self.config.talker_config.text_config.vocab_size,
                )
                if i not in (self.config.talker_config.codec_eos_token_id,)
            ]  # Suppress additional special tokens, should not be predicted
            talker_kwargs = {
                "max_new_tokens": talker_max_new_tokens,
                "do_sample": talker_temperature > 0.0,
                "top_k": talker_top_k,
                "top_p": talker_top_p,
                "temperature": talker_temperature,
                "eos_token_id": self.config.talker_config.codec_eos_token_id,
                "repetition_penalty": talker_repetition_penalty,
                "suppress_tokens": talker_supppressed_tokens,
                "output_hidden_states": True,
                "return_dict_in_generate": True,
            }
            for thinker_chunk in thinker_chunk_stream:
                yield {
                    "thinker_ids": thinker_chunk["thinker_generate_ids"],
                }
                if (
                    thinker_chunk["thinker_generate_ids"].shape[-1] == 1
                    and thinker_chunk["thinker_generate_ids"][0, -1].item() in skip_token_ids
                ):
                    # Skip if only one token generated, which is likely to be a stop token
                    continue
                thinker_generate_hidden_states = thinker_chunk["thinker_generate_hidden_states"]
                full_generated_ids = thinker_chunk["full_generated_ids"]
                wav_iter = self.talker_generate(
                    hidden_states=thinker_generate_hidden_states,
                    sequences=full_generated_ids,
                    input_ids=input_ids,
                    speaker_id=speaker_id,
                    talker_kwargs=talker_kwargs,
                    token2wav_kwargs=token2wav_kwargs,
                )  # [1,1,T]
                for talker_wav in wav_iter:
                    yield {
                        "talker_wav": talker_wav.float(),  # BFloat16 -> Float32
                    }

    @torch.no_grad()
    def generate(
        self,
        input_ids: Optional[torch.Tensor] = None,
        speaker: str = "Ethan",
        use_audio_in_video: bool = False,
        return_audio: Optional[bool] = None,
        thinker_max_new_tokens: int = 1024,
        thinker_eos_token_id: int = 151645,
        talker_max_new_tokens: int = 4096,
        talker_do_sample: bool = True,
        talker_top_k: int = 50,
        talker_top_p: float = 1.0,
        talker_temperature: float = 0.9,
        talker_repetition_penalty: float = 1.05,
        **kwargs,
    ):
        if return_audio and not self.has_talker:
            raise ValueError(
                "Cannot use talker when talker module not initialized. Use `enable_talker` method or set enable_talker in config to enable talker."
            )
        if return_audio is None:
            return_audio = self.has_talker

        shared_kwargs = {"use_audio_in_video": use_audio_in_video}
        thinker_kwargs = {
            "max_new_tokens": thinker_max_new_tokens,
            "eos_token_id": thinker_eos_token_id,
        }

        talker_kwargs = {}
        token2wav_kwargs = {}
        if return_audio:
            # "chelsie": 2301, "ethan": 2302, "aiden": 2303
            speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
            if speaker_id is None:
                raise NotImplementedError(f"Speaker {speaker} not implemented")
            if input_ids.shape[0] != 1:
                raise NotImplementedError(
                    "Qwen3-Omni currently does not support batched inference with audio output"
                )
            talker_supppressed_tokens = [
                i
                for i in range(
                    self.config.talker_config.text_config.vocab_size - 1024,
                    self.config.talker_config.text_config.vocab_size,
                )
                if i not in (self.config.talker_config.codec_eos_token_id,)
            ]  # Suppress additional special tokens, should not be predicted
            talker_kwargs = {
                "max_new_tokens": talker_max_new_tokens,
                "do_sample": talker_do_sample,
                "top_k": talker_top_k,
                "top_p": talker_top_p,
                "temperature": talker_temperature,
                "eos_token_id": self.config.talker_config.codec_eos_token_id,
                "repetition_penalty": talker_repetition_penalty,
                "suppress_tokens": talker_supppressed_tokens,
                "output_hidden_states": True,
                "return_dict_in_generate": True,
            }
            token2wav_kwargs = {}

        for key, value in kwargs.items():
            if key.startswith("thinker_"):
                thinker_kwargs[key[len("thinker_") :]] = value
            elif key.startswith("talker_"):
                talker_kwargs[key[len("talker_") :]] = value
            elif key.startswith("token2wav_"):
                token2wav_kwargs[key[len("token2wav_") :]] = value
            # Process special input values
            elif key == "feature_attention_mask":
                thinker_kwargs[key] = value
                talker_kwargs["audio_feature_lengths"] = torch.sum(value, dim=1)
            elif key in ("input_features", "attention_mask"):
                thinker_kwargs[key] = value
            # Put other key to shared kwargs
            else:
                shared_kwargs[key] = value

        # Merge kwargs
        for key, value in shared_kwargs.items():
            if key not in thinker_kwargs:
                thinker_kwargs[key] = value
            if key not in talker_kwargs and key in [
                "image_grid_thw",
                "video_grid_thw",
                "video_second_per_grid",
            ]:
                talker_kwargs[key] = value
            if key not in token2wav_kwargs:
                token2wav_kwargs[key] = value

        # 1. Generate from thinker module
        generate_audio = return_audio and self.has_talker
        if generate_audio:
            thinker_kwargs["output_hidden_states"] = True
            thinker_kwargs["return_dict_in_generate"] = True

        thinker_result = self.thinker.generate(input_ids=input_ids, **thinker_kwargs)

        if not generate_audio:
            return thinker_result, None

        # logging.info(f"talker input sequences: {thinker_result.sequences}")
        # logging.info(f"hidden_states tuples size {len(thinker_result.hidden_states)}")
        # for i, item in enumerate(thinker_result.hidden_states):
        #    if isinstance(item, tuple):
        #        logging.info(
        #            f"hidden_states tuple len: {len(item)}, item shape:{item[0].shape}",
        #        )
        #    elif isinstance(item, torch.Tensor):
        #        logging.info(f"{i=}", item.shape)
        #    else:
        #        logging.info(f"{i=}", item)

        talker_wavs = self.talker_generate(
            hidden_states=thinker_result.hidden_states,
            sequences=thinker_result.sequences,
            input_ids=input_ids,
            speaker_id=speaker_id,
            talker_kwargs=talker_kwargs,
            token2wav_kwargs=token2wav_kwargs,
        )  # [1,1,T]
        return thinker_result, talker_wavs



import os
import sys
import asyncio

import modal

CONFIG_PATH = os.getenv("CONFIG_PATH", "config/chat_with_minicpm_gs.yaml")

image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    # https://hub.docker.com/r/pytorch/pytorch/tags
    modal.Image.from_registry(
        "pytorch/pytorch:2.4.1-cuda12.1-cudnn9-devel",
        add_python="3.11",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang")
    .env(
        {
            "USE_GPTQ_CKPT": os.getenv("USE_GPTQ_CKPT", ""),
            # https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#gpu-feature-list
            # nvcc --list-gpu-arch
            "TORCH_CUDA_ARCH_LIST": "7.5;8.0;8.6+PTX;8.9+PTX;9.0",  # for auto-gptq install
        }
    )
    .run_commands(
        "which nvcc",
        "nvcc --version",
        "pip install --no-build-isolation achatbot[flash-attn]",
    )
    .run_commands(
        [
            "if [ $USE_GPTQ_CKPT ];then"
            + " git clone https://github.com/OpenBMB/AutoGPTQ.git -b minicpmo"
            + " && cd AutoGPTQ"
            + " && pip install -vvv --no-build-isolation -e . "
            + ";fi",
        ]
    )
    .pip_install("uv")
    .run_commands(
        "pip install chumpy==0.70 --no-build-isolation",
        "pip install mmcv==2.2.0 -f https://download.openmmlab.com/mmcv/dist/cu121/torch2.4/index.html",
    )
    .run_commands(
        "which python",
        "python --version",
        "git lfs install --skip-smudge",
        "git clone https://github.com/weedge/OpenAvatarChat.git -b feat/achatbot",
        "cd /OpenAvatarChat && git submodule foreach git lfs pull",
        "cd /OpenAvatarChat && git lfs install --force",
        "cd /OpenAvatarChat && git checkout 468ae5d9d6b1481b68ee94d3decdc8efcafbd105",
        "cd /OpenAvatarChat && git submodule update --init --recursive;git lfs logs last",
    )
    .run_commands(
        "cd /OpenAvatarChat && git pull origin feat/achatbot && git checkout 136725008748e69c47d3444f0b02bc658269bd26",
        f"cd /OpenAvatarChat && python install.py --uv --config {CONFIG_PATH}",
        "rm -rf /OpenAvatarChat/models",
        "rm -rf /OpenAvatarChat/resource",
    )
    .env(
        {
            "CONFIG_PATH": CONFIG_PATH,
        }
    )
    .pip_install("twilio", "pydantic==2.11.7")
)


# ----------------------- app -------------------------------
app = modal.App("open_avatar_chat_bot")


HF_MODEL_DIR = "/OpenAvatarChat/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
RESOURCES_DIR = "/OpenAvatarChat/resource"
resources_vol = modal.Volume.from_name("resources", create_if_missing=True)
TORCH_CACHE_DIR = "/root/.cache/torch"
torch_cache_vol = modal.Volume.from_name("torch_cache", create_if_missing=True)

"""
# download model weights
modal run src/download_models.py --repo-ids "FunAudioLLM/SenseVoiceSmall"
modal run src/download_models.py --repo-ids "openbmb/MiniCPM-o-2_6,openbmb/MiniCPM-o-2_6-int4" 
modal run src/download_models.py --repo-ids "facebook/wav2vec2-base-960h"
modal run src/download_models.py::download_ckpts --ckpt-urls "https://huggingface.co/3DAIGC/LAM_audio2exp/resolve/main/LAM_audio2exp_streaming.tar"

IMAGE_GPU=T4 CONFIG_PATH=config/echo_with_gs.yaml modal serve src/avatar/open_avatar_chat.py
IMAGE_GPU=T4 CONFIG_PATH=config/chat_with_gs.yaml modal serve src/avatar/open_avatar_chat.py

IMAGE_GPU=L4 CONFIG_PATH=config/chat_with_minicpm_gs.yaml modal serve src/avatar/open_avatar_chat.py
IMAGE_GPU=T4 USE_GPTQ_CKPT=1 CONFIG_PATH=config/chat_with_minicpm_int4_gs.yaml modal serve src/avatar/open_avatar_chat.py

"""


# RTC_CONFIG = {"iceServers": [{"url": "stun:stun.l.google.com:19302"}]}


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.function(
    image=image,
    gpu=os.getenv("IMAGE_GPU", "T4"),
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        RESOURCES_DIR: resources_vol,
        TORCH_CACHE_DIR: torch_cache_vol,
    },
    secrets=[modal.Secret.from_name("achatbot")],
    cpu=4.0,
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
    # NOTE: don't use max_inputs in function params
    # max_inputs=int(os.getenv("IMAGE_CONCURRENT_CN", "1")),
)
@modal.concurrent(max_inputs=int(os.getenv("IMAGE_CONCURRENT_CN", "10")))  # inputs per container
@modal.asgi_app()  # ASGI on Modal
def ui():
    import subprocess
    import torch

    subprocess.run("nvidia-smi --version", shell=True)
    gpu_prop = None
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda:0")
        print(gpu_prop)
        torch.multiprocessing.set_start_method("spawn", force=True)
    else:
        print("CUDA is not available.")

    return init_app()


def init_app():
    import gradio as gr
    import uvicorn
    from fastapi import FastAPI
    from fastapi.responses import RedirectResponse
    from gradio.routes import mount_gradio_app  # connects Gradio and FastAPI

    sys.path.insert(0, "/OpenAvatarChat/src")

    from engine_utils.directory_info import DirectoryInfo
    from service.service_utils.service_config_loader import load_configs
    from service.service_utils.logger_utils import config_loggers
    from chat_engine.chat_engine import ChatEngine

    project_dir = DirectoryInfo.get_project_dir()
    if project_dir not in sys.path:
        sys.path.insert(0, project_dir)

    logger_config, service_config, engine_config = load_configs(os.getenv("CONFIG_PATH"))

    config_loggers(logger_config)

    # set up demo
    app = FastAPI()

    @app.get("/")
    def get_root():
        return RedirectResponse(url="/ui")

    @app.get("/ui/static/fonts/system-ui/system-ui-Regular.woff2")
    @app.get("/ui/static/fonts/ui-sans-serif/ui-sans-serif-Regular.woff2")
    @app.get("/favicon.ico")
    def get_font():
        # remove confusing error
        return {}

    css = """
    .app {
        @media screen and (max-width: 768px) {
            padding: 8px !important;
        }
    }
    footer {
        display: none !important;
    }
    """
    with gr.Blocks(css=css) as gradio_block:
        with gr.Column():
            with gr.Group() as rtc_container:
                pass
    app = gr.mount_gradio_app(app, gradio_block, "/ui")

    chat_engine = ChatEngine()
    chat_engine.initialize(engine_config, app=app, ui=gradio_block, parent_block=rtc_container)

    return app



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""LSTM layers module."""

from torch import nn


class SLSTM(nn.Module):
    """
    LSTM without worrying about the hidden state, nor the layout of the data.
    Expects input as convolutional layout.
    """
    def __init__(self, dimension: int, num_layers: int = 2, skip: bool = True):
        super().__init__()
        self.skip = skip
        self.lstm = nn.LSTM(dimension, dimension, num_layers)

    # def forward(self, x):
    #     x = x.permute(2, 0, 1)
    #     y, _ = self.lstm(x)
    #     if self.skip:
    #         y = y + x
    #     y = y.permute(1, 2, 0)
    #     return y

    # 修改transpose顺序
    def forward(self, x):
        # # 插入reshape
        # x = x.reshape(x.shape)
        x1 = x.permute(2, 0, 1)
        y, _ = self.lstm(x1)
        y = y.permute(1, 2, 0)
        if self.skip:
            y = y + x
        return y



from fish_speech.i18n import i18n

HEADER_MD = f"""# Fish Speech

{i18n("A text-to-speech model based on VQ-GAN and Llama developed by [Fish Audio](https://fish.audio).")}  

{i18n("You can find the source code [here](https://github.com/fishaudio/fish-speech) and models [here](https://huggingface.co/fishaudio/fish-speech-1.5).")}  

{i18n("Related code and weights are released under CC BY-NC-SA 4.0 License.")}  

{i18n("We are not responsible for any misuse of the model, please consider your local laws and regulations before using it.")}  
"""

TEXTBOX_PLACEHOLDER = i18n("Put your text here.")



import io
import logging
from threading import Thread
from PIL import Image
from time import perf_counter

try:
    from transformers import AutoTokenizer, TextIteratorStreamer, AutoModel
    import torch

except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Deepseek-OCR, you need to `pip install achatbot[llm_transformers_manual_vision_deepseek_ocr]`"
    )
    raise Exception(f"Missing module: {e}")


from achatbot.common.utils.helper import get_device, print_model_params
from achatbot.common.interface import IVisionOCR
from achatbot.common.random import set_all_random_seed
from achatbot.common.session import Session
from achatbot.types.speech.language import TO_LLM_LANGUAGE
from achatbot.types.llm.transformers import TransformersLMArgs
from .base import TransformersBaseLLM


class TransformersManualVisionDeepSeekOCR(TransformersBaseLLM, IVisionOCR):
    TAG = "llm_transformers_manual_vision_deepseek_ocr"

    def __init__(self, tokenizer=None, **args) -> None:
        self.base_size = args.pop("ocr_base_size", 1024)
        self.image_size = args.pop("ocr_image_size", 640)
        self.crop_mode = args.pop("ocr_crop_mode", True)
        # Tiny: base_size = 512, image_size = 512, crop_mode = False
        # Small: base_size = 640, image_size = 640, crop_mode = False
        # Base: base_size = 1024, image_size = 1024, crop_mode = False
        # Large: base_size = 1280, image_size = 1280, crop_mode = False
        # Gundam: base_size = 1024, image_size = 640, crop_mode = True # default

        self.prompt = args.pop(
            "ocr_prompt", "<image>\n<|grounding|>Convert the document to markdown. "
        )
        # document: <image>\n<|grounding|>Convert the document to markdown.
        # other image: <image>\n<|grounding|>OCR this image.
        # without layouts: <image>\nFree OCR.
        # figures in document: <image>\nParse the figure.
        # general: <image>\nDescribe this image in detail.
        # rec: <image>\nLocate <|ref|>xxxx<|/ref|> in the image.

        self.args = TransformersLMArgs(**args)
        gpu_prop = torch.cuda.get_device_properties("cuda") if torch.cuda.is_available() else None

        if self.args.lm_device_map:
            self._model = AutoModel.from_pretrained(
                self.args.lm_model_name_or_path,
                torch_dtype=torch.bfloat16,
                #!NOTE: https://github.com/huggingface/transformers/issues/20896
                # device_map for multi cpu/gpu with accelerate
                device_map=self.args.lm_device_map,
                attn_implementation="flash_attention_2"
                if gpu_prop and gpu_prop.major >= 8
                else None,
                trust_remote_code=True,
            ).eval()
        else:
            self._model = (
                AutoModel.from_pretrained(
                    self.args.lm_model_name_or_path,
                    torch_dtype=torch.bfloat16,
                    attn_implementation="flash_attention_2"
                    if gpu_prop and gpu_prop.major >= 8
                    else None,
                    trust_remote_code=True,
                )
                .eval()
                .to(self.args.lm_device)
            )

        logging.info(f"TransformersLMArgs: {self.args}")
        print_model_params(self._model, self.TAG)
        self._tokenizer = tokenizer or AutoTokenizer.from_pretrained(
            self.args.lm_model_name_or_path, use_fast=True
        )
        self.warmup()

    def warmup(self):
        if self.args.warmup_steps <= 0:
            return

        # create dummy image file
        dummy_img = Image.new("RGB", (640, 640), color="white")
        ioBuff = io.BytesIO()
        dummy_img.save(ioBuff, format="PNG")
        ioBuff.seek(0)

        streamer = TextIteratorStreamer(
            self._tokenizer, skip_prompt=True, skip_special_tokens=False
        )

        warmup_gen_kwargs = dict(
            tokenizer=self._tokenizer,
            prompt=self.prompt,
            image_file=ioBuff,
            base_size=self.base_size,
            image_size=self.image_size,
            crop_mode=self.crop_mode,
            save_results=False,
            test_compress=False,
            eval_mode=False,
            verbose=False,
            streamer=streamer,
        )

        self._warmup(
            target=self._model.infer,
            kwargs=warmup_gen_kwargs,
            streamer=streamer,
        )

    def set_task_prompt(self, prompt: str):
        self.prompt = prompt

    @torch.inference_mode()
    async def async_generate(self, session: Session, **kwargs):
        seed = kwargs.get("seed", self.args.lm_gen_seed)
        set_all_random_seed(seed)

        ocr_img = session.ctx.state["ocr_img"]
        streamer = TextIteratorStreamer(
            self._tokenizer, skip_prompt=True, skip_special_tokens=False
        )
        generation_kwargs = dict(
            tokenizer=self._tokenizer,
            prompt=self.prompt,
            image_file=ocr_img,
            base_size=self.base_size,
            image_size=self.image_size,
            crop_mode=self.crop_mode,
            save_results=False,
            test_compress=False,
            eval_mode=False,
            verbose=False,
            streamer=streamer,
        )
        thread = Thread(target=self._model.infer, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        start = perf_counter()
        times = []
        is_ref_det = False
        sentence = ""
        # TODO: extract detect image to storage with s3 (use callback)
        for new_text in streamer:
            times.append(perf_counter() - start)
            generated_text += new_text
            if "<|ref|>" in new_text:
                is_ref_det = True

            if "<|/det|>" in new_text:
                is_ref_det = False
                new_text = new_text.split("<|/det|>")[1]

            if "<｜end▁of▁sentence｜>" in new_text:
                if "<|/ref|>" not in new_text:
                    new_text = new_text.split("<｜end▁of▁sentence｜>")[0]
            if "<|end▁of▁sentence|>" in new_text:
                if "<|/ref|>" not in new_text:
                    new_text = new_text.split("<|end▁of▁sentence|>")[0]

            if is_ref_det is False:
                sentence += new_text
                pos = self._have_special_char(sentence)
                if pos > -1:
                    yield sentence[: pos + 1]
                    sentence = sentence[pos + 1 :]
            start = perf_counter()
        if len(sentence) > 0:
            yield sentence + "."
        if times:
            logging.info(f"{generated_text=} TTFT: {times[0]:.4f}s total time: {sum(times):.4f}s")
        torch.cuda.empty_cache()



# -*- coding: utf-8 -*-
"""
https://github.com/Xiaobin-Rong/gtcrn
"""

import torch


def convert_to_stream(stream_model, model):
    state_dict = model.state_dict()
    new_state_dict = stream_model.state_dict()

    for key in stream_model.state_dict().keys():
        if key in state_dict.keys():
            new_state_dict[key] = state_dict[key]

        elif key.replace("Conv1d.", "") in state_dict.keys():
            new_state_dict[key] = state_dict[key.replace("Conv1d.", "")]

        elif key.replace("Conv2d.", "") in state_dict.keys():
            new_state_dict[key] = state_dict[key.replace("Conv2d.", "")]

        ## For StreamConvTranspose2d Version 1:
        # elif key.replace('ConvTranspose2d.', '') in state_dict.keys():
        #     new_state_dict[key] = state_dict[key.replace('ConvTranspose2d.', '')]

        ## For StreamConvTranspose2d Version 2:
        elif key.replace("ConvTranspose2d.", "") in state_dict.keys():
            if key.endswith("weight"):
                if (
                    new_state_dict[key].shape
                    != state_dict[key.replace("ConvTranspose2d.", "")].shape
                ):
                    new_state_dict[key] = torch.flip(
                        state_dict[key.replace("ConvTranspose2d.", "")].permute([1, 0, 2, 3]),
                        dims=[-2, -1],
                    )
                else:
                    new_state_dict[key] = torch.flip(
                        state_dict[key.replace("ConvTranspose2d.", "")], dims=[-2, -1]
                    )

            else:
                new_state_dict[key] = state_dict[key.replace("ConvTranspose2d.", "")]

        else:
            raise (ValueError("key error!"))

    stream_model.load_state_dict(new_state_dict)



import numpy as np
import torch
from librosa.filters import mel as librosa_mel_fn
from scipy.io.wavfile import read

MAX_WAV_VALUE = 32768.0


def load_wav(full_path):
    sampling_rate, data = read(full_path)
    return data, sampling_rate


def dynamic_range_compression(x, C=1, clip_val=1e-5):
    return np.log(np.clip(x, a_min=clip_val, a_max=None) * C)


def dynamic_range_decompression(x, C=1):
    return np.exp(x) / C


def dynamic_range_compression_torch(x, C=1, clip_val=1e-5):
    return torch.log(torch.clamp(x, min=clip_val) * C)


def dynamic_range_decompression_torch(x, C=1):
    return torch.exp(x) / C


def spectral_normalize_torch(magnitudes):
    output = dynamic_range_compression_torch(magnitudes)
    return output


def spectral_de_normalize_torch(magnitudes):
    output = dynamic_range_decompression_torch(magnitudes)
    return output


mel_basis = {}
hann_window = {}


def mel_spectrogram(y, n_fft=1920, num_mels=80, sampling_rate=24000, hop_size=480,
                    win_size=1920, fmin=0, fmax=8000, center=False):
    global mel_basis, hann_window  # pylint: disable=global-statement
    if f"{str(fmax)}_{str(y.device)}" not in mel_basis:
        mel = librosa_mel_fn(sr=sampling_rate, n_fft=n_fft, n_mels=num_mels, fmin=fmin, fmax=fmax)
        mel_basis[str(fmax) + "_" + str(y.device)] = torch.from_numpy(mel).float().to(y.device)
        hann_window[str(y.device)] = torch.hann_window(win_size).to(y.device)

    y = torch.nn.functional.pad(
        y.unsqueeze(1), (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)), mode="reflect"
    )
    y = y.squeeze(1)

    spec = torch.view_as_real(
        torch.stft(
            y,
            n_fft,
            hop_length=hop_size,
            win_length=win_size,
            window=hann_window[str(y.device)],
            center=center,
            pad_mode="reflect",
            normalized=False,
            onesided=True,
            return_complex=True,
        )
    )

    spec = torch.sqrt(spec.pow(2).sum(-1) + (1e-9))

    spec = torch.matmul(mel_basis[str(fmax) + "_" + str(y.device)], spec)
    spec = spectral_normalize_torch(spec)

    return spec



import os
import logging

import random
import unittest

from src.modules.speech.tts import TTSEnvInit
from src.common.interface import ITts
from src.common.factory import EngineClass
from src.common.logger import Logger
from src.common.session import Session
from src.common.types import SessionCtx

r"""
python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_get_voices
python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_set_voice
python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_synthesize

TTS_TAG=tts_onnx_kokoro python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_get_voices
TTS_TAG=tts_onnx_kokoro python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_set_voice
TTS_TAG=tts_onnx_kokoro KOKORO_ESPEAK_NG_LIB_PATH=/usr/local/lib/libespeak-ng.1.dylib python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_synthesize
TTS_STREAM=1 TTS_TAG=tts_onnx_kokoro KOKORO_ESPEAK_NG_LIB_PATH=/usr/local/lib/libespeak-ng.1.dylib python -m unittest test.modules.speech.tts.test_kokoro.TestKOKORO.test_synthesize
"""


class TestKOKORO(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tts_tag = os.getenv("TTS_TAG", "tts_kokoro")
        cls.tts_text = os.getenv(
            "TTS_TEXT",
            # "hello, test.modules.speech.tts.test_kokoro.TestKOKORO.test_synthesize.",
            "你好，我是机器人",
            # "How could I know? It's an unanswerable question. Like asking an unborn child if they'll lead a good life. They haven't even been born.",
        )
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.tts: EngineClass | ITts = TTSEnvInit.initTTSEngine(self.tts_tag)
        self.session = Session(**SessionCtx("test_tts_client_id").__dict__)

    def tearDown(self):
        pass

    def test_get_voices(self):
        voices = self.tts.get_voices()
        self.assertGreater(len(voices), 0)
        print(voices)

    def test_set_voice(self):
        voice = random.choice(self.tts.get_voices())
        self.tts.set_voice(voice)

        print(voice)
        self.assertEqual(self.tts.args.voice, voice)

    def test_synthesize(self):
        import pyaudio

        stream_info = self.tts.get_stream_info()
        print(stream_info)
        self.pyaudio_instance = pyaudio.PyAudio()
        self.audio_stream = self.pyaudio_instance.open(
            format=stream_info["format"],
            channels=stream_info["channels"],
            rate=stream_info["rate"],
            output_device_index=None,
            output=True,
        )

        # self.test_set_voice()

        self.session.ctx.state["tts_text"] = self.tts_text
        print(self.session.ctx)
        iter = self.tts.synthesize_sync(self.session)
        sub_chunk_size = 1024
        for i, chunk in enumerate(iter):
            print(f"get {i} chunk {len(chunk)}")
            self.assertGreaterEqual(len(chunk), 0)
            if len(chunk) / sub_chunk_size < 100:
                self.audio_stream.write(chunk)
                continue
            for i in range(0, len(chunk), sub_chunk_size):
                sub_chunk = chunk[i : i + sub_chunk_size]
                self.audio_stream.write(sub_chunk)

        self.audio_stream.stop_stream()
        self.audio_stream.close()
        self.pyaudio_instance.terminate()



import os
import time
import threading
import queue
import logging
from io import BytesIO

from tqdm import tqdm
from scipy.interpolate import interp1d
from pydub import AudioSegment
from pydub.silence import detect_silence
import soundfile as sf
import wave
import cv2
import torch
from torchvision import transforms
import numpy as np


def geneHeadInfo(sampleRate, bits, sampleNum):
    import struct

    rHeadInfo = b"\x52\x49\x46\x46"
    fileLength = struct.pack("i", sampleNum + 36)
    rHeadInfo += fileLength
    rHeadInfo += b"\x57\x41\x56\x45\x66\x6d\x74\x20\x10\x00\x00\x00\x01\x00\x01\x00"
    rHeadInfo += struct.pack("i", sampleRate)
    rHeadInfo += struct.pack("i", int(sampleRate * bits / 8))
    rHeadInfo += b"\x02\x00"
    rHeadInfo += struct.pack("H", bits)
    rHeadInfo += b"\x64\x61\x74\x61"
    rHeadInfo += struct.pack("i", sampleNum)
    return rHeadInfo


class liteAvatar(object):
    def __init__(
        self,
        data_dir=None,  # avatar data dir (mouth pos)
        num_threads=1,
        use_bg_as_idle=False,
        fps=30,
        generate_offline=False,
        use_gpu=False,
        weight_dir="./weights",
    ):
        logging.info("liteAvatar init start...")

        self.data_dir = data_dir
        self.fps = fps
        self.use_bg_as_idle = use_bg_as_idle
        self.use_gpu = use_gpu
        self.device = "cuda" if use_gpu else "cpu"

        s = time.time()
        from audio2mouth import Audio2Mouth

        self.audio2mouth = Audio2Mouth(use_gpu=use_gpu, weight_dir=weight_dir)
        logging.info(f"audio2mouth init over in {time.time() - s}s")

        self.input_queue = queue.Queue()
        self.output_queue = queue.Queue()
        self.load_data_thread: threading.Thread = None

        logging.info("liteAvatar init over")
        self._generate_offline = generate_offline
        if generate_offline:
            self.load_dynamic_model(data_dir)

            self.threads_prep = []
            barrier_prep = threading.Barrier(num_threads, action=None, timeout=None)
            for i in range(num_threads):
                t = threading.Thread(
                    target=self.face_gen_loop,
                    args=(i, barrier_prep, self.input_queue, self.output_queue),
                )
                self.threads_prep.append(t)

            for t in self.threads_prep:
                t.daemon = True
                t.start()

    def stop_algo(self):
        pass

    def load_dynamic_model(self, data_dir):
        logging.info("start to load dynamic data")
        start_time = time.time()
        self.encoder = torch.jit.load(f"{data_dir}/net_encode.pt").to(self.device)
        self.generator = torch.jit.load(f"{data_dir}/net_decode.pt").to(self.device)

        self.load_pos_bg_video_image_frames(data_dir=data_dir, bg_frame_cnt=150)
        self.load_ref_mouth_images(data_dir=data_dir, bg_frame_cnt=150)
        self.ref_data_list = [0 for x in range(150)]
        self.input_queue = queue.Queue()
        self.output_queue = queue.Queue()
        logging.info(f"load dynamic model in {time.time() - start_time:.3f}s")

    def unload_dynamic_model(self):
        pass

    def load_pos_bg_video_image_frames(self, data_dir, bg_frame_cnt=None):
        """load pos np.ndarray and bg video image frames"""
        t = time.time()
        self.neutral_pose = np.load(f"{data_dir}/neutral_pose.npy")
        self.mouth_scale = None

        self.bg_data_list = []
        bg_video = cv2.VideoCapture(f"{data_dir}/bg_video.mp4")
        while True:
            ret, img = bg_video.read()
            self.bg_data_list.append(img)
            if ret is False:
                break
        self.bg_video_frame_count = (
            len(self.bg_data_list)
            if bg_frame_cnt is None
            else min(bg_frame_cnt, len(self.bg_data_list))
        )

        y1, y2, x1, x2 = open(f"{data_dir}/face_box.txt", "r").readlines()[0].split()
        self.y1, self.y2, self.x1, self.x2 = int(y1), int(y2), int(x1), int(x2)

        self.merge_mask = (np.ones((self.y2 - self.y1, self.x2 - self.x1, 3)) * 255).astype(
            np.uint8
        )
        self.merge_mask[10:-10, 10:-10, :] *= 0
        self.merge_mask = cv2.GaussianBlur(self.merge_mask, (21, 21), 15)
        self.merge_mask = self.merge_mask / 255

        self.frame_vid_list = []

        self.image_transforms = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize([0.5], [0.5]),
            ]
        )
        logging.info(f"load data sync in {time.time() - t:.3f}s")

    def load_ref_mouth_images(self, data_dir, bg_frame_cnt=None):
        """load ref mouth images from bg_video.mp4"""
        logging.info(f"loading data from {data_dir}")
        s = time.time()

        self.ref_img_list = []
        for ii in tqdm(range(bg_frame_cnt)):
            img_file_path = os.path.join(f"{data_dir}", "ref_frames", f"ref_{ii:05d}.jpg")
            image = cv2.cvtColor(cv2.imread(img_file_path)[:, :, 0:3], cv2.COLOR_BGR2RGB)
            image = cv2.resize(image, (384, 384), interpolation=cv2.INTER_LINEAR)
            ref_img = self.image_transforms(np.uint8(image))
            encoder_input = ref_img.unsqueeze(0).float().to(self.device)
            x = self.encoder(encoder_input)
            self.ref_img_list.append(x)

        logging.info(f"load data over in {time.time() - s}s")

    def face_gen_loop(self, thread_id, barrier, in_queue, out_queue):
        while True:
            try:
                data = in_queue.get()
            except queue.Empty:
                break

            if data is None:
                in_queue.put(None)
                break

            s = time.time()

            param_res = data[0]
            bg_frame_id = data[1]
            global_frame_id = data[2]

            mouth_img = self.param2img(param_res, bg_frame_id)
            full_img, mouth_img = self.merge_mouth_to_bg(mouth_img, bg_frame_id)

            logging.info(
                "global_frame_id: {} in {}s".format(global_frame_id, round(time.time() - s, 3))
            )

            out_queue.put((global_frame_id, full_img, mouth_img))

        barrier.wait()
        if thread_id == 0:
            out_queue.put(None)

    def param2img(
        self, param_res: dict, bg_frame_id: int, global_frame_id=0, is_idle=False
    ) -> torch.Tensor:
        param_val = []
        for key in param_res:
            val = param_res[key]
            param_val.append(val)
        param_val = np.asarray(param_val)

        source_img = self.generator(
            self.ref_img_list[bg_frame_id],
            torch.from_numpy(param_val).unsqueeze(0).float().to(self.device),
        )
        source_img = source_img.detach().to("cpu")

        return source_img

    def get_idle_param(self):
        bg_param = self.neutral_pose
        tmp_json = {}
        for ii in range(len(self.audio2mouth.p_list)):
            tmp_json[str(ii)] = float(bg_param[ii])
        return tmp_json

    def merge_mouth_to_bg(
        self, mouth_image: torch.Tensor, bg_frame_id: int, use_bg: bool = False
    ) -> tuple[np.ndarray, np.ndarray]:
        mouth_image = (mouth_image / 2 + 0.5).clamp(0, 1)
        mouth_image = mouth_image[0].permute(1, 2, 0) * 255

        mouth_image = mouth_image.numpy().astype(np.uint8)
        mouth_image = cv2.resize(mouth_image, (self.x2 - self.x1, self.y2 - self.y1))
        mouth_image = mouth_image[:, :, ::-1]
        full_img = self.bg_data_list[bg_frame_id].copy()
        if not use_bg:
            # image no bg, use default bg
            full_img[self.y1 : self.y2, self.x1 : self.x2, :] = (
                mouth_image * (1 - self.merge_mask)
                + full_img[self.y1 : self.y2, self.x1 : self.x2, :] * self.merge_mask
            )
        full_img = full_img.astype(np.uint8)
        return full_img, mouth_image.astype(np.uint8)

    def interp_param(self, param_res, fps=25):
        old_len = len(param_res)
        new_len = int(old_len / 30 * fps + 0.5)

        interp_list = {}
        for key in param_res[0]:
            tmp_list = []
            for ii in range(len(param_res)):
                tmp_list.append(param_res[ii][key])
            tmp_list = np.asarray(tmp_list)

            x = np.linspace(0, old_len - 1, num=old_len, endpoint=True)
            newx = np.linspace(0, old_len - 1, num=new_len, endpoint=True)
            f = interp1d(x, tmp_list)
            y = f(newx)
            interp_list[key] = y

        new_param_res = []
        for ii in range(new_len):
            tmp_json = {}
            for key in interp_list:
                tmp_json[key] = interp_list[key][ii]
            new_param_res.append(tmp_json)

        return new_param_res

    def padding_last(self, param_res, last_end=None):
        bg_param = self.neutral_pose

        if last_end is None:
            last_end = len(param_res)

        padding_cnt = 5
        final_end = max(last_end + 5, len(param_res))
        param_res = param_res[:last_end]
        padding_list = []
        for ii in range(last_end, final_end):
            tmp_json = {}
            for key in param_res[-1]:
                kk = ii - last_end
                scale = max((padding_cnt - kk - 1) / padding_cnt, 0.0)

                end_value = bg_param[int(key)]
                tmp_json[key] = (param_res[-1][key] - end_value) * scale + end_value
            padding_list.append(tmp_json)

        logging.info(f"padding_cnt:{len(padding_list)}")
        param_res = param_res + padding_list
        return param_res

    def audio2param(
        self, input_audio_byte, prefix_padding_size=0, is_complete=False, audio_status=-1
    ):
        headinfo = geneHeadInfo(16000, 16, len(input_audio_byte))
        input_audio_byte = headinfo + input_audio_byte
        input_audio, sr = sf.read(BytesIO(input_audio_byte))

        param_res, _, _ = self.audio2mouth.inference(subtitles=None, input_audio=input_audio)

        sil_scale = np.zeros(len(param_res))
        sound = AudioSegment.from_raw(
            BytesIO(input_audio_byte), sample_width=2, frame_rate=16000, channels=1
        )
        start_end_list = detect_silence(sound, 500, -50)
        if len(start_end_list) > 0:
            for start, end in start_end_list:
                start_frame = int(start / 1000 * 30)
                end_frame = int(end / 1000 * 30)
                logging.info(f"silence part: {start_frame}-{end_frame} frames")
                sil_scale[start_frame:end_frame] = 1
        sil_scale = np.pad(sil_scale, 2, mode="edge")
        kernel = np.array([0.1, 0.2, 0.4, 0.2, 0.1])
        sil_scale = np.convolve(sil_scale, kernel, "same")
        sil_scale = sil_scale[2:-2]
        self.make_silence(param_res, sil_scale)
        if self.fps != 30:
            param_res = self.interp_param(param_res, fps=self.fps)

        if is_complete:
            param_res = self.padding_last(param_res)

        return param_res

    def make_silence(self, param_res, sil_scale):
        bg_param = self.neutral_pose

        for ii in range(len(param_res)):
            for key in param_res[ii]:
                neu_value = bg_param[int(key)]
                param_res[ii][key] = (
                    param_res[ii][key] * (1 - sil_scale[ii]) + neu_value * sil_scale[ii]
                )
        return param_res

    def handle(self, audio_file_path, result_dir, param_res=None):
        audio_data = self.read_wav_to_bytes(audio_file_path)

        if param_res is None:
            param_res = self.audio2param(audio_data)

        for ii in range(len(param_res)):
            s = time.time()
            frame_id = ii
            if int(frame_id / self.bg_video_frame_count) % 2 == 0:
                frame_id = frame_id % self.bg_video_frame_count
            else:
                frame_id = self.bg_video_frame_count - 1 - frame_id % self.bg_video_frame_count
            self.input_queue.put((param_res[ii], frame_id, ii))

        self.input_queue.put(None)

        tmp_frame_dir = os.path.join(result_dir, "tmp_frames")
        if os.path.exists(tmp_frame_dir):
            os.system(f"rm -rf {tmp_frame_dir}")
        os.mkdir(tmp_frame_dir)

        while True:
            res_data = self.output_queue.get()
            if res_data is None:
                break
            global_frame_index = res_data[0]
            target_path = f"{tmp_frame_dir}/{str(global_frame_index + 1).zfill(5)}.jpg"
            cv2.imwrite(target_path, res_data[1])

        for p in self.threads_prep:
            p.join()

        cmd = "`which ffmpeg` -r 30 -i {}/%05d.jpg -i {} -framerate 30 -c:v libx264 -pix_fmt yuv420p -b:v 5000k -strict experimental -loglevel error {}/test_demo.mp4 -y".format(
            tmp_frame_dir, audio_file_path, result_dir
        )
        os.system(cmd)

    @staticmethod
    def read_wav_to_bytes(file_path):
        try:
            # 打开WAV文件
            with wave.open(file_path, "rb") as wav_file:
                # 获取WAV文件的参数
                params = wav_file.getparams()
                logging.info(
                    f"Channels: {params.nchannels}, Sample Width: {params.sampwidth}, Frame Rate: {params.framerate}, Number of Frames: {params.nframes}"
                )

                # 读取所有帧
                frames = wav_file.readframes(params.nframes)
                return frames
        except wave.Error as e:
            logging.warning(f"Error reading WAV file: {e}")
            return None


"""
huggingface-cli download weege007/liteavatar --local-dir ./models/weege007/liteavatar
modelscope download --model HumanAIGC-Engineering/LiteAvatarGallery 20250408/sample_data.zip --local_dir ./resources/avatar/liteavatar
unzip ./resources/avatar/liteavatar/20250408/sample_data.zip -d ./resources/avatar/liteavatar/20250408
PYTHONPATH=$PYTHONPATH:./deps/LiteAvatar python -m deps.LiteAvatar.lite_avatar \
    --weight_dir ./models/weege007/liteavatar \
    --data_dir ./resources/avatar/liteavatar/20250408/sample_data \
    --audio_file ./test/audio_files/asr_example_zh.wav \
    --result_dir ./

modelscope download --model HumanAIGC-Engineering/LiteAvatarGallery 20250612/P1-64AzfrJY037WpS69RiUMw.zip --local_dir ./resources/avatar/liteavatar
unzip ./resources/avatar/liteavatar/20250612/P1-64AzfrJY037WpS69RiUMw.zip -d ./resources/avatar/liteavatar/20250612
PYTHONPATH=$PYTHONPATH:./deps/LiteAvatar python -m deps.LiteAvatar.lite_avatar \
    --weight_dir ./models/weege007/liteavatar \
    --data_dir ./resources/avatar/liteavatar/20250612/P1-64AzfrJY037WpS69RiUMw \
    --audio_file ./test/audio_files/asr_example_zh.wav \
    --result_dir ./
"""
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str)
    parser.add_argument("--audio_file", type=str)
    parser.add_argument("--result_dir", type=str)
    parser.add_argument("--weight_dir", type=str)
    parser.add_argument("--log_level", type=str, default="INFO")
    args = parser.parse_args()

    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s"
    logging.basicConfig(
        level=args.log_level,
        format=log_format,
    )

    audio_file = args.audio_file
    tmp_frame_dir = args.result_dir
    weight_dir = args.weight_dir
    use_gpu = True if torch.cuda.is_available() else False

    lite_avatar = liteAvatar(
        data_dir=args.data_dir,
        num_threads=1,
        generate_offline=True,
        weight_dir=weight_dir,
        use_gpu=use_gpu,
    )

    lite_avatar.handle(audio_file, tmp_frame_dir)



"""
# Copyright 2024-2025 The Alibaba 3DAIGC Team Authors. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from engines.defaults import (
    default_argument_parser,
    default_config_parser,
    default_setup,
)
from engines.infer import INFER
from engines.launch import launch


def main_worker(cfg):
    cfg = default_setup(cfg)
    infer = INFER.build(dict(type=cfg.infer.type, cfg=cfg))
    infer.infer()


def main():
    args = default_argument_parser().parse_args()
    cfg = default_config_parser(args.config_file, args.options)

    launch(
        main_worker,
        num_gpus_per_machine=args.num_gpus,
        num_machines=args.num_machines,
        machine_rank=args.machine_rank,
        dist_url=args.dist_url,
        cfg=(cfg,),
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2019 Shigeki Karita
#               2020 Mobvoi Inc (Binbin Zhang)
#               2022 Xingchen Song (sxc19@mails.tsinghua.edu.cn)
#               2024 Alibaba Inc (Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multi-Head Attention layer definition."""

import math
from typing import Tuple

import torch
from torch import nn


class MultiHeadedAttention(nn.Module):
    """Multi-Head Attention layer.

    Args:
        n_head (int): The number of heads.
        n_feat (int): The number of features.
        dropout_rate (float): Dropout rate.

    """

    def __init__(self,
                 n_head: int,
                 n_feat: int,
                 dropout_rate: float,
                 key_bias: bool = True):
        """Construct an MultiHeadedAttention object."""
        super().__init__()
        assert n_feat % n_head == 0
        # We assume d_v always equals d_k
        self.d_k = n_feat // n_head
        self.h = n_head
        self.linear_q = nn.Linear(n_feat, n_feat)
        self.linear_k = nn.Linear(n_feat, n_feat, bias=key_bias)
        self.linear_v = nn.Linear(n_feat, n_feat)
        self.linear_out = nn.Linear(n_feat, n_feat)
        self.dropout = nn.Dropout(p=dropout_rate)

    def forward_qkv(
        self, query: torch.Tensor, key: torch.Tensor, value: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Transform query, key and value.

        Args:
            query (torch.Tensor): Query tensor (#batch, time1, size).
            key (torch.Tensor): Key tensor (#batch, time2, size).
            value (torch.Tensor): Value tensor (#batch, time2, size).

        Returns:
            torch.Tensor: Transformed query tensor, size
                (#batch, n_head, time1, d_k).
            torch.Tensor: Transformed key tensor, size
                (#batch, n_head, time2, d_k).
            torch.Tensor: Transformed value tensor, size
                (#batch, n_head, time2, d_k).

        """
        n_batch = query.size(0)
        q = self.linear_q(query).view(n_batch, -1, self.h, self.d_k)
        k = self.linear_k(key).view(n_batch, -1, self.h, self.d_k)
        v = self.linear_v(value).view(n_batch, -1, self.h, self.d_k)
        q = q.transpose(1, 2)  # (batch, head, time1, d_k)
        k = k.transpose(1, 2)  # (batch, head, time2, d_k)
        v = v.transpose(1, 2)  # (batch, head, time2, d_k)

        return q, k, v

    def forward_attention(
        self,
        value: torch.Tensor,
        scores: torch.Tensor,
        mask: torch.Tensor = torch.ones((0, 0, 0), dtype=torch.bool)
    ) -> torch.Tensor:
        """Compute attention context vector.

        Args:
            value (torch.Tensor): Transformed value, size
                (#batch, n_head, time2, d_k).
            scores (torch.Tensor): Attention score, size
                (#batch, n_head, time1, time2).
            mask (torch.Tensor): Mask, size (#batch, 1, time2) or
                (#batch, time1, time2), (0, 0, 0) means fake mask.

        Returns:
            torch.Tensor: Transformed value (#batch, time1, d_model)
                weighted by the attention score (#batch, time1, time2).

        """
        n_batch = value.size(0)
        # NOTE(xcsong): When will `if mask.size(2) > 0` be True?
        #   1. onnx(16/4) [WHY? Because we feed real cache & real mask for the
        #           1st chunk to ease the onnx export.]
        #   2. pytorch training
        if mask.size(2) > 0:  # time2 > 0
            mask = mask.unsqueeze(1).eq(0)  # (batch, 1, *, time2)
            # For last chunk, time2 might be larger than scores.size(-1)
            mask = mask[:, :, :, :scores.size(-1)]  # (batch, 1, *, time2)
            scores = scores.masked_fill(mask, -float('inf'))
            attn = torch.softmax(scores, dim=-1).masked_fill(
                mask, 0.0)  # (batch, head, time1, time2)
        # NOTE(xcsong): When will `if mask.size(2) > 0` be False?
        #   1. onnx(16/-1, -1/-1, 16/0)
        #   2. jit (16/-1, -1/-1, 16/0, 16/4)
        else:
            attn = torch.softmax(scores, dim=-1)  # (batch, head, time1, time2)

        p_attn = self.dropout(attn)
        x = torch.matmul(p_attn, value)  # (batch, head, time1, d_k)
        x = (x.transpose(1, 2).contiguous().view(n_batch, -1,
                                                 self.h * self.d_k)
             )  # (batch, time1, d_model)

        return self.linear_out(x)  # (batch, time1, d_model)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor = torch.ones((0, 0, 0), dtype=torch.bool),
        pos_emb: torch.Tensor = torch.empty(0),
        cache: torch.Tensor = torch.zeros((0, 0, 0, 0))
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute scaled dot product attention.

        Args:
            query (torch.Tensor): Query tensor (#batch, time1, size).
            key (torch.Tensor): Key tensor (#batch, time2, size).
            value (torch.Tensor): Value tensor (#batch, time2, size).
            mask (torch.Tensor): Mask tensor (#batch, 1, time2) or
                (#batch, time1, time2).
                1.When applying cross attention between decoder and encoder,
                the batch padding mask for input is in (#batch, 1, T) shape.
                2.When applying self attention of encoder,
                the mask is in (#batch, T, T)  shape.
                3.When applying self attention of decoder,
                the mask is in (#batch, L, L)  shape.
                4.If the different position in decoder see different block
                of the encoder, such as Mocha, the passed in mask could be
                in (#batch, L, T) shape. But there is no such case in current
                CosyVoice.
            cache (torch.Tensor): Cache tensor (1, head, cache_t, d_k * 2),
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`


        Returns:
            torch.Tensor: Output tensor (#batch, time1, d_model).
            torch.Tensor: Cache tensor (1, head, cache_t + time1, d_k * 2)
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`

        """
        q, k, v = self.forward_qkv(query, key, value)

        # NOTE(xcsong):
        #   when export onnx model, for 1st chunk, we feed
        #       cache(1, head, 0, d_k * 2) (16/-1, -1/-1, 16/0 mode)
        #       or cache(1, head, real_cache_t, d_k * 2) (16/4 mode).
        #       In all modes, `if cache.size(0) > 0` will alwayse be `True`
        #       and we will always do splitting and
        #       concatnation(this will simplify onnx export). Note that
        #       it's OK to concat & split zero-shaped tensors(see code below).
        #   when export jit  model, for 1st chunk, we always feed
        #       cache(0, 0, 0, 0) since jit supports dynamic if-branch.
        # >>> a = torch.ones((1, 2, 0, 4))
        # >>> b = torch.ones((1, 2, 3, 4))
        # >>> c = torch.cat((a, b), dim=2)
        # >>> torch.equal(b, c)        # True
        # >>> d = torch.split(a, 2, dim=-1)
        # >>> torch.equal(d[0], d[1])  # True
        if cache.size(0) > 0:
            key_cache, value_cache = torch.split(cache,
                                                 cache.size(-1) // 2,
                                                 dim=-1)
            k = torch.cat([key_cache, k], dim=2)
            v = torch.cat([value_cache, v], dim=2)
        # NOTE(xcsong): We do cache slicing in encoder.forward_chunk, since it's
        #   non-trivial to calculate `next_cache_start` here.
        new_cache = torch.cat((k, v), dim=-1)

        scores = torch.matmul(q, k.transpose(-2, -1)) / math.sqrt(self.d_k)
        return self.forward_attention(v, scores, mask), new_cache


class RelPositionMultiHeadedAttention(MultiHeadedAttention):
    """Multi-Head Attention layer with relative position encoding.
    Paper: https://arxiv.org/abs/1901.02860
    Args:
        n_head (int): The number of heads.
        n_feat (int): The number of features.
        dropout_rate (float): Dropout rate.
    """

    def __init__(self,
                 n_head: int,
                 n_feat: int,
                 dropout_rate: float,
                 key_bias: bool = True):
        """Construct an RelPositionMultiHeadedAttention object."""
        super().__init__(n_head, n_feat, dropout_rate, key_bias)
        # linear transformation for positional encoding
        self.linear_pos = nn.Linear(n_feat, n_feat, bias=False)
        # these two learnable bias are used in matrix c and matrix d
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        self.pos_bias_u = nn.Parameter(torch.Tensor(self.h, self.d_k))
        self.pos_bias_v = nn.Parameter(torch.Tensor(self.h, self.d_k))
        torch.nn.init.xavier_uniform_(self.pos_bias_u)
        torch.nn.init.xavier_uniform_(self.pos_bias_v)

    def rel_shift(self, x: torch.Tensor) -> torch.Tensor:
        """Compute relative positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, head, time1, 2*time1-1).
            time1 means the length of query vector.

        Returns:
            torch.Tensor: Output tensor.

        """
        zero_pad = torch.zeros((x.size()[0], x.size()[1], x.size()[2], 1),
                               device=x.device,
                               dtype=x.dtype)
        x_padded = torch.cat([zero_pad, x], dim=-1)

        x_padded = x_padded.view(x.size()[0],
                                 x.size()[1],
                                 x.size(3) + 1, x.size(2))
        x = x_padded[:, :, 1:].view_as(x)[
            :, :, :, : x.size(-1) // 2 + 1
        ]  # only keep the positions from 0 to time2
        return x

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor = torch.ones((0, 0, 0), dtype=torch.bool),
        pos_emb: torch.Tensor = torch.empty(0),
        cache: torch.Tensor = torch.zeros((0, 0, 0, 0)),
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute 'Scaled Dot Product Attention' with rel. positional encoding.
        Args:
            query (torch.Tensor): Query tensor (#batch, time1, size).
            key (torch.Tensor): Key tensor (#batch, time2, size).
            value (torch.Tensor): Value tensor (#batch, time2, size).
            mask (torch.Tensor): Mask tensor (#batch, 1, time2) or
                (#batch, time1, time2), (0, 0, 0) means fake mask.
            pos_emb (torch.Tensor): Positional embedding tensor
                (#batch, time2, size).
            cache (torch.Tensor): Cache tensor (1, head, cache_t, d_k * 2),
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`
        Returns:
            torch.Tensor: Output tensor (#batch, time1, d_model).
            torch.Tensor: Cache tensor (1, head, cache_t + time1, d_k * 2)
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`
        """
        q, k, v = self.forward_qkv(query, key, value)
        q = q.transpose(1, 2)  # (batch, time1, head, d_k)

        # NOTE(xcsong):
        #   when export onnx model, for 1st chunk, we feed
        #       cache(1, head, 0, d_k * 2) (16/-1, -1/-1, 16/0 mode)
        #       or cache(1, head, real_cache_t, d_k * 2) (16/4 mode).
        #       In all modes, `if cache.size(0) > 0` will alwayse be `True`
        #       and we will always do splitting and
        #       concatnation(this will simplify onnx export). Note that
        #       it's OK to concat & split zero-shaped tensors(see code below).
        #   when export jit  model, for 1st chunk, we always feed
        #       cache(0, 0, 0, 0) since jit supports dynamic if-branch.
        # >>> a = torch.ones((1, 2, 0, 4))
        # >>> b = torch.ones((1, 2, 3, 4))
        # >>> c = torch.cat((a, b), dim=2)
        # >>> torch.equal(b, c)        # True
        # >>> d = torch.split(a, 2, dim=-1)
        # >>> torch.equal(d[0], d[1])  # True
        if cache is not None and cache.size(0) > 0:
            key_cache, value_cache = torch.split(cache, cache.size(-1) // 2, dim=-1)
            k = torch.cat([key_cache, k], dim=2)
            v = torch.cat([value_cache, v], dim=2)
        # NOTE(xcsong): We do cache slicing in encoder.forward_chunk, since it's
        #   non-trivial to calculate `next_cache_start` here.
        new_cache = torch.cat((k, v), dim=-1)

        n_batch_pos = pos_emb.size(0)
        p = self.linear_pos(pos_emb).view(n_batch_pos, -1, self.h, self.d_k)
        p = p.transpose(1, 2)  # (batch, head, time1, d_k)

        # (batch, head, time1, d_k)
        q_with_bias_u = (q + self.pos_bias_u).transpose(1, 2)
        # (batch, head, time1, d_k)
        q_with_bias_v = (q + self.pos_bias_v).transpose(1, 2)

        # compute attention score
        # first compute matrix a and matrix c
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        # (batch, head, time1, time2)
        matrix_ac = torch.matmul(q_with_bias_u, k.transpose(-2, -1))

        # compute matrix b and matrix d
        # (batch, head, time1, time2)
        matrix_bd = torch.matmul(q_with_bias_v, p.transpose(-2, -1))
        # NOTE(Xiang Lyu): Keep rel_shift since espnet rel_pos_emb is used
        if matrix_ac.shape != matrix_bd.shape:
            matrix_bd = self.rel_shift(matrix_bd)

        scores = (matrix_ac + matrix_bd) / math.sqrt(
            self.d_k)  # (batch, head, time1, time2)

        return self.forward_attention(v, scores, mask), new_cache



import logging

from dotenv import load_dotenv
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger
from apipeline.frames import AudioRawFrame, TextFrame

from src.processors.speech.audio_save_processor import AudioSaveProcessor
from src.processors.filters.audio_noise_filter import AudioNoiseFilter
from src.processors.aggregators.user_audio_response import UserAudioResponseAggregator
from src.cmd.bots.base_livekit import LivekitRoomBot
from src.common.types import LivekitParams
from src.transports.livekit import LivekitTransport
from src.cmd.bots import register_ai_room_bots
from src.modules.speech.enhancer import SpeechEnhancerEnvInit


load_dotenv(override=True)


@register_ai_room_bots.register
class LivekitAudioNoiseFilterBot(LivekitRoomBot):
    """
    filter audio noise to echo
    - use livekit audio stream(bytes) --> audio noise filter processor --> save
    """

    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

        self.vad_analyzer = None
        self.audio_noise_filter = None
        self.se_args = {}

    def load(self):
        self.vad_analyzer = self.get_vad_analyzer()

        # SE engine
        se_config = self.bot_config().se
        self.se_args = {} if se_config.args is None else se_config.args
        self.audio_noise_filter = SpeechEnhancerEnvInit.initEngine(se_config.tag, **self.se_args)
        self.audio_noise_filter.warmup(self.session)

    async def arun(self):
        assert self.vad_analyzer is not None
        assert self.audio_noise_filter is not None

        self.params = LivekitParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            vad_enabled=True,
            vad_analyzer=self.vad_analyzer,
            vad_audio_passthrough=True,
        )
        logging.info(f"params: {self.params}")

        transport = LivekitTransport(
            self.args.token,
            params=self.params,
        )

        audio_noise_filter = AudioNoiseFilter(
            se=self.audio_noise_filter, session=self.session, **self.se_args
        )
        user_audio_save_processor = None
        if self._save_audio:
            user_audio_save_processor = AudioSaveProcessor(
                prefix_name="user_speak", pass_raw_audio=True
            )

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    audio_noise_filter,
                    # UserAudioResponseAggregator(),
                    # user_audio_save_processor,
                    # FrameLogger(include_frame_types=[AudioRawFrame]),
                    transport.output_processor(),  # BotSpeakingFrame
                ]
            ),
            params=PipelineParams(
                allow_interruptions=False,
                enable_metrics=False,
                send_initial_empty_metrics=False,
            ),
        )

        transport.add_event_handler("on_first_participant_joined", self.on_first_participant_joined)

        await PipelineRunner().run(self.task)



import torch.nn as nn
import torch
import torch.nn.functional as F
import copy


class MlpProjector(nn.Module):
    def __init__(self, cfg):
        super().__init__()

        self.cfg = cfg

        if cfg.projector_type == "identity":
            modules = nn.Identity()

        elif cfg.projector_type == "linear":
            modules = nn.Linear(cfg.input_dim, cfg.n_embed)

        elif cfg.projector_type == "mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            modules = [nn.Linear(cfg.input_dim, cfg.n_embed)]
            for _ in range(1, mlp_depth):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed, cfg.n_embed))
            modules = nn.Sequential(*modules)

        elif cfg.projector_type == "normlayer_downsample_mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            mlp_ratio = cfg.get("mlp_ratio", 1)
            modules = [
                nn.LayerNorm(cfg.input_dim * cfg.downsample_ratio * cfg.downsample_ratio),
                nn.Linear(
                    cfg.input_dim * cfg.downsample_ratio * cfg.downsample_ratio,
                    cfg.n_embed * mlp_ratio,
                ),
            ]
            for _ in range(1, mlp_depth - 1):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed * mlp_ratio, cfg.n_embed * mlp_ratio))
            modules.append(nn.GELU())
            modules.append(nn.Linear(cfg.n_embed * mlp_ratio, cfg.n_embed))
            modules = nn.Sequential(*modules)

        elif cfg.projector_type == "downsample_mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            mlp_ratio = cfg.get("mlp_ratio", 1)
            modules = [
                nn.Linear(
                    cfg.input_dim * cfg.downsample_ratio * cfg.downsample_ratio,
                    cfg.n_embed * mlp_ratio,
                )
            ]
            for _ in range(1, mlp_depth - 1):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed * mlp_ratio, cfg.n_embed * mlp_ratio))
            modules.append(nn.GELU())
            modules.append(nn.Linear(cfg.n_embed * mlp_ratio, cfg.n_embed))
            modules = nn.Sequential(*modules)

        elif cfg.projector_type == "low_high_hybrid_split_mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            self.high_up_proj = nn.Linear(cfg.input_dim, cfg.n_embed // 2)
            self.low_up_proj = nn.Linear(cfg.input_dim, cfg.n_embed // 2)

            modules = []
            for _ in range(1, mlp_depth):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed, cfg.n_embed))
            modules = nn.Sequential(*modules)

        elif cfg.projector_type == "hybrid_split_feature_mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            channel_div = cfg.get("channel_div", 0.5)
            self.high_up_proj = nn.Linear(cfg.input_dim[0], int(cfg.n_embed * channel_div))
            self.low_up_proj = nn.Linear(
                cfg.input_dim[1], cfg.n_embed - int(cfg.n_embed * channel_div)
            )

            modules = []
            for _ in range(1, mlp_depth):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed, cfg.n_embed))
            modules = nn.Sequential(*modules)

        elif cfg.projector_type == "low_high_split_mlp_gelu":
            mlp_depth = cfg.get("depth", 1)
            modules = []
            for _ in range(1, mlp_depth):
                modules.append(nn.GELU())
                modules.append(nn.Linear(cfg.n_embed // 2, cfg.n_embed // 2))
            modules = nn.Sequential(*modules)
            self.high_layers = nn.Sequential(*modules)
            self.low_layers = copy.deepcopy(modules)

        else:
            raise ValueError(f"Unknown projector type: {cfg.projector_type}")

        if cfg.get("token_pooling", False):
            self.token_pooling_layer = nn.Linear(cfg.input_dim * 4, cfg.input_dim)

        if cfg.get("conv_fusion_high_low_features", False):
            self.fusion_layer = nn.Linear(cfg.input_dim, cfg.input_dim)
        self.layers = modules

    def forward(self, x):
        if self.cfg.get("token_pooling", False):
            batch_size, wxh, channels = x.shape
            w = h = int(wxh**0.5)
            x = x.view(batch_size, w, h, channels)
            x = x.permute(0, 3, 1, 2)
            # import ipdb; ipdb.set_trace()
            patches = x.unfold(2, 2, 2).unfold(3, 2, 2)
            batch_size, channels, h_patches, w_patches, _, _ = patches.size()
            # 在通道维度上拼接
            patches = patches.contiguous().view(batch_size, channels, h_patches * w_patches, -1)

            # 通过线性层
            patches = patches.permute(0, 2, 1, 3).contiguous()
            patches = patches.view(batch_size, h_patches * w_patches, channels * 4)

            x = self.token_pooling_layer(patches)

        if self.cfg.get("conv_fusion_high_low_features", False):
            x = self.fusion_layer(x[:, 0]) + x[:, 1]

        if self.cfg.projector_type == "low_high_hybrid_split_mlp_gelu":
            high_x, low_x = x[0], x[1]
            high_x = self.high_up_proj(high_x)
            low_x = self.low_up_proj(low_x)
            x = torch.concat([high_x, low_x], dim=-1)

        if self.cfg.projector_type == "hybrid_split_feature_mlp_gelu":
            high_x = x[..., : self.cfg.input_dim[0]]
            low_x = x[..., self.cfg.input_dim[0] :]
            high_x = self.high_up_proj(high_x)
            low_x = self.low_up_proj(low_x)
            x = torch.concat([high_x, low_x], dim=-1)

        if self.cfg.projector_type == "low_high_split_mlp_gelu":
            high_x, low_x = x[0], x[1]
            high_x = self.high_layers(high_x)
            low_x = self.low_layers(low_x)
            x = torch.concat([high_x, low_x], dim=-1)
            return x

        if (
            self.cfg.projector_type == "downsample_mlp_gelu"
            or self.cfg.projector_type == "normlayer_downsample_mlp_gelu"
        ):
            bs, hw, input_dim = x.shape
            h = w = int((hw) ** 0.5)

            """compute padding"""
            if h % self.cfg.downsample_ratio:
                pad = self.cfg.downsample_ratio - h % self.cfg.downsample_ratio
            else:
                pad = 0
            x = x.reshape(bs, h, w, input_dim)
            if pad > 0:
                x = F.pad(x, (0, 0, 0, pad, 0, pad), "constant", 0)

            """4 to 1 concat"""
            x = x.permute(0, 3, 1, 2)  # B, C, H, W
            x = F.unfold(
                x,
                kernel_size=self.cfg.downsample_ratio,
                stride=self.cfg.downsample_ratio,
                padding=0,
            )  # B, C*4, HW // 4
            x = x.permute(0, 2, 1)

        return self.layers(x)

    @staticmethod
    def get_flops_per_sample(cfg):
        if cfg.projector_type == "linear":
            fwd = 2 * cfg.input_dim * cfg.n_embed

        elif "mlp_gelu" in cfg.projector_type:
            mlp_depth = cfg.get("depth", 1)
            downsample_ratio = cfg.get("downsample_ratio", 1)
            input_dim = sum(cfg.input_dim) if isinstance(cfg.input_dim, list) else cfg.input_dim
            input_dim = input_dim * downsample_ratio * downsample_ratio
            fwd = 2 * input_dim * cfg.n_embed + (mlp_depth - 1) * 2 * cfg.n_embed * cfg.n_embed
        else:
            fwd = 0

        return fwd * 3



# ASR
def asr_test(model):
    messages = [
        "请记录下你所听到的语音内容。",
        {"type": "audio", "audio": "assets/give_me_a_brief_introduction_to_the_great_wall.wav"},
    ]
    tokens, text, _ = model(messages, max_new_tokens=256, temperature=0.1, do_sample=True)
    print(text)

# S2TT（support: en,zh,ja）
def s2tt_test(model):
    messages = [
        "请仔细聆听这段语音，然后将其内容翻译成中文",
        # "Please listen carefully to this audio and then translate its content into Chinese.",
        {"type": "audio", "audio": "assets/give_me_a_brief_introduction_to_the_great_wall.wav"},
    ]
    tokens, text, _ = model(messages, max_new_tokens=256, temperature=0.1, do_sample=True)
    print(text)


# audio caption
def audio_caption_test(model):
    messages = [
        "Please briefly explain the important events involved in this audio clip.",
        {"type": "audio", "audio": "assets/music_playing_followed_by_a_woman_speaking.wav"},
    ]
    tokens, text, _ = model(messages, max_new_tokens=256, temperature=0.1, do_sample=True)
    print(text)

# TTS（support: en,zh,ja)
def tts_test(model, token2wav):
    messages = [
        "以自然的语速读出下面的文字。\n",
        # "Read this paragraph at a natural pace.\n",
        "你好呀，我是你的AI助手，很高兴认识你！"
        "<tts_start>"
    ]
    tokens, text, audio = model(messages, max_tokens=2048, temperature=0.7, do_sample=True)
    print(text)
    #print(tokens)
    audio = [x for x in audio if x < 6561] # remove audio padding
    audio = token2wav(audio, prompt_wav='assets/default_male.wav')
    with open('output-tts.wav', 'wb') as f:
        f.write(audio)

# T2ST（support: en,zh,ja)
def t2st_test(model, token2wav):
    messages = [
        "将下面的文本翻译成英文，并用语音播报。\n",
        # "Translate the following text into English and broadcast it with voice.\n",
        "你好呀，我是你的AI助手，很高兴认识你！"
        "<tts_start>"
    ]
    tokens, text, audio = model(messages, max_tokens=2048, temperature=0.7, do_sample=True)
    print(text)
    #print(tokens)
    audio = [x for x in audio if x < 6561] # remove audio padding
    audio = token2wav(audio, prompt_wav='assets/default_male.wav')
    with open('output-t2st.wav', 'wb') as f:
        f.write(audio)

# S2ST（support: en,zh）
def s2st_test(model, token2wav):
    messages = [
        "请仔细聆听这段语音，然后将其内容翻译成中文并用语音播报。",
        # "Please listen carefully to this audio and then translate its content into Chinese speech.",
        {"type": "audio", "audio": "assets/give_me_a_brief_introduction_to_the_great_wall.wav"},
        "<tts_start>"
    ]
    tokens, text, audio = model(messages, max_tokens=2048, temperature=0.7, do_sample=True)
    print(text)
    #print(tokens)
    audio = [x for x in audio if x < 6561] # remove audio padding
    audio = token2wav(audio, prompt_wav='assets/default_female.wav')
    with open('output-s2st.wav', 'wb') as f:
        f.write(audio)

# multi turn aqta
def multi_turn_aqta_test(model):
    history = []
    for round_idx, inp_audio in enumerate([
        "assets/multi-turn-round1-听说荡口古镇从下个月开始取消门票了，你知道这事吗。.wav",
        "assets/multi-turn-round2-新闻说九月十九号就免费开放了。好像整个古镇都升级改造了，现在变成开放式街区了。.wav"
    ]):
        print("round: ", round_idx)
        history.append(
            {"type": "audio", "audio": inp_audio}
        )
        tokens, text, _ = model(history, max_new_tokens=256, temperature=0.5, do_sample=True)
        print(text)
        history.append(text)

# multi turn aqaa
def multi_turn_aqaa_test(model, token2wav):
    history = []
    for round_idx, inp_audio in enumerate([
        "assets/multi-turn-round1-听说荡口古镇从下个月开始取消门票了，你知道这事吗。.wav",
        "assets/multi-turn-round2-新闻说九月十九号就免费开放了。好像整个古镇都升级改造了，现在变成开放式街区了。.wav"
    ]):
        print("round: ", round_idx)
        history.append(
            {"type": "audio", "audio": inp_audio},
        )
        history.append("<tts_start>")
        tokens, text, audio = model(history, max_new_tokens=2048, temperature=0.7, do_sample=True)
        print(text)
        audio = [x for x in audio if x < 6561] # remove audio padding
        audio = token2wav(audio, prompt_wav='assets/default_female.wav')
        with open(f'output-round-{round_idx}.wav', 'wb') as f:
            f.write(audio)
        history.append(
            {"type":"token", "token": tokens}
        )

if __name__ == '__main__':
    from stepaudio2 import StepAudio2Base
    from token2wav import Token2wav

    model = StepAudio2Base('Step-Audio-2-mini-Base')
    token2wav = Token2wav('Step-Audio-2-mini-Base/token2wav')
    asr_test(model)
    s2tt_test(model)
    audio_caption_test(model)
    tts_test(model, token2wav)
    t2st_test(model, token2wav)
    s2st_test(model, token2wav)
    multi_turn_aqta_test(model)
    multi_turn_aqaa_test(model, token2wav)



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from typing import Optional, Tuple

import torch

from transformers import AutoConfig, AutoModelForCausalLM, \
    MptConfig, MptForCausalLM, MptModel
from llava.model.llava_arch import LlavaMetaModel, LlavaMetaForCausalLM


class LlavaMptConfig(MptConfig):
    model_type = "llava_mpt"


class LlavaMptModel(LlavaMetaModel, MptModel):
    config_class = LlavaMptConfig

    def __init__(self, config: MptConfig):
        config.hidden_size = config.d_model
        super(LlavaMptModel, self).__init__(config)

    def embed_tokens(self, x):
        return self.wte(x)


class LlavaMptForCausalLM(MptForCausalLM, LlavaMetaForCausalLM):
    config_class = LlavaMptConfig
    supports_gradient_checkpointing = True

    def __init__(self, config):
        super(MptForCausalLM, self).__init__(config)

        self.transformer = LlavaMptModel(config)
        self.lm_head = torch.nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_model(self):
        return self.transformer

    def _set_gradient_checkpointing(self, module, value=False):
        if isinstance(module, LlavaMptModel):
            module.gradient_checkpointing = value

    def forward(
            self,
            input_ids: Optional[torch.LongTensor] = None,
            past_key_values: Optional[Tuple[Tuple[torch.Tensor, torch.Tensor], ...]] = None,
            attention_mask: Optional[torch.Tensor] = None,
            inputs_embeds: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None,
            use_cache: Optional[bool] = None,
            output_attentions: Optional[bool] = None,
            output_hidden_states: Optional[bool] = None,
            return_dict: Optional[bool] = None,
            images=None):

        input_ids, attention_mask, past_key_values, inputs_embeds, labels = self.prepare_inputs_labels_for_multimodal(input_ids, attention_mask, past_key_values, labels, images)

        return super().forward(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            labels=labels,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

    def prepare_inputs_for_generation(self, input_ids, past_key_values=None, inputs_embeds=None, **kwargs):
        images = kwargs.pop("images", None)
        _inputs = super().prepare_inputs_for_generation(
            input_ids, past_key_values=past_key_values, inputs_embeds=inputs_embeds, **kwargs
        )
        _inputs['images'] = images
        return _inputs


AutoConfig.register("llava_mpt", LlavaMptConfig)
AutoModelForCausalLM.register(LlavaMptConfig, LlavaMptForCausalLM)



import asyncio
import logging
import os
import argparse
from contextlib import asynccontextmanager
import traceback
from typing import Dict
import json
from asyncio import TimeoutError


from fastapi import FastAPI, WebSocket, BackgroundTasks
from fastapi.middleware.cors import CORSMiddleware
from fastapi import WebSocketDisconnect
from dotenv import load_dotenv

from src.common.utils.thread_safe import ThreadSafeDict
from src.cmd.bots.bridge.base import AISmallWebRTCFastapiWebsocketBot
from src.cmd.bots.base import AIBot
from src.cmd.bots.bot_loader import BotLoader
from src.common.types import CONFIG_DIR
from src.common.const import *
from src.common.logger import Logger
from src.cmd.http.server.fastapi_daily_bot_serve import ngrok_proxy
from src.services.webrtc_peer_connection import SmallWebRTCConnection, IceServer


load_dotenv(override=True)
Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

run_bot: AISmallWebRTCFastapiWebsocketBot = None
config = None

# TODO: connect session mrg
# Store websocket connection
# ws_map = ThreadSafeDict()
ws_map: Dict[str, WebSocket] = {}
# Store rtc connections
# pcs_map = ThreadSafeDict()
pcs_map: Dict[str, SmallWebRTCConnection] = {}
# Store peer rtc connection pending candidates
# pending_candidates = ThreadSafeDict()
pending_candidates: Dict[str, list] = {}

ice_servers = [
    IceServer(
        urls="stun:stun.l.google.com:19302",
    )
]


# https://fastapi.tiangolo.com/advanced/events/#lifespan
@asynccontextmanager
async def lifespan(app: FastAPI):
    global run_bot
    try:
        # load model before running
        config_file = config.f if config else os.getenv("CONFIG_FILE")
        run_bot = await BotLoader.load_bot(config_file, bot_type="fastapi_ws_bot")
        run_bot.load()
    except Exception as e:
        print(e)
        traceback.print_exc()
        return

    print(f"load bot {run_bot} success")

    yield  # Run app

    # app life end to clear resources
    # clear websocket connection
    coros = [ws.close() for ws in ws_map.values() if ws.state == "OPEN"]
    await asyncio.gather(*coros)
    ws_map.clear()
    print(f"websocket connections clear success")

    # clear webrtc connection
    coros = [pc.disconnect() for pc in pcs_map.values() if pc.connectionState == "connected"]
    await asyncio.gather(*coros)
    pcs_map.clear()
    pending_candidates.clear()
    print(f"rtc peer connections clear success")


app = FastAPI(lifespan=lifespan)

# 配置CORS中间件
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # 允许所有来源，生产环境中应该限制为特定域名
    allow_credentials=True,  # 允许携带凭证
    allow_methods=["*"],  # 允许所有HTTP方法
    allow_headers=["*"],  # 允许所有HTTP头部
)


@app.post("/api/ice_candidate/{peer_id}")
async def handle_ice_candidate(candidate: dict, peer_id: str):
    """
    post remote peer ice candidate
    """
    if not candidate:
        return

    logging.info(f"received ice candidate from {peer_id}: {candidate.get('candidate_sdp')}...")

    if not pending_candidates.get(peer_id):
        pending_candidates[peer_id] = []

    if not pcs_map.get(peer_id):
        pending_candidates[peer_id].append(candidate)
    else:
        if len(pending_candidates[peer_id]) > 0:
            await asyncio.gather(
                *(pcs_map[peer_id].add_ice_candidate(c) for c in pending_candidates[peer_id])
            )
            pending_candidates[peer_id] = []
        await pcs_map[peer_id].add_ice_candidate(candidate)


@app.post("/api/offer/{peer_id}")
async def handle_offer(request: dict, background_tasks: BackgroundTasks, peer_id: str):
    logging.info(f"request: {request} peer_id: {peer_id}")

    if not peer_id:
        logging.error(f"peer_id is empty")
        return None

    if peer_id and peer_id in pcs_map:
        connection = pcs_map[peer_id]
        logging.info(f"Reusing existing connection for peer_id: {peer_id}")
        await connection.renegotiate(
            sdp=request.get("sdp"),
            type=request.get("type"),
            restart_pc=request.get("restart_pc", False),
        )
    else:
        connection = SmallWebRTCConnection(ice_servers)
        await connection.initialize(
            sdp=request.get("sdp"),
            type=request.get("type"),
        )

        @connection.event_handler("closed")
        async def handle_disconnected(webrtc_connection: SmallWebRTCConnection):
            logging.info(f"Discarding peer connection for pc_id: {peer_id}")

            # Remove from pcs_map
            pcs_map.pop(peer_id, None)

            # Close associated websocket if it exists
            if peer_id in ws_map:
                try:
                    ws = ws_map[peer_id]
                    if ws.state == "OPEN":
                        await ws.close(code=1000, reason="WebRTC connection closed")
                    ws_map.pop(peer_id, None)
                    logging.info(
                        f"Closed websocket for peer_id: {peer_id} due to WebRTC disconnection"
                    )
                except Exception as e:
                    logging.error(f"Error closing websocket for {peer_id}: {str(e)}")

    answer = connection.get_answer()
    logging.info(f"answer bot peer_id: {answer.get('pc_id')}")
    pcs_map[peer_id] = connection

    return answer


# 添加心跳超时设置
HEARTBEAT_TIMEOUT = 35  # 35 seconds (slightly longer than client's 30-second interval)


@app.websocket("/{peer_id}")
async def websocket_endpoint(websocket: WebSocket, peer_id: str):
    try:
        if not peer_id:
            logging.error(f"pc_id is empty")
            return
        if peer_id and peer_id not in pcs_map:
            logging.error(f"{peer_id=} not found webrtc peer connection")
            return

        await websocket.accept()
        ws_map[peer_id] = websocket
        logging.info(f"accept {peer_id=} client: {websocket.client}")

        # Set up the WebRTC connection
        run_bot.set_webrtc_connection(pcs_map[peer_id])
        run_bot.set_fastapi_websocket(websocket)

        # Start the main bot task
        bot_task = asyncio.create_task(run_bot.async_run())

        # Start the heartbeat monitoring loop
        last_heartbeat = asyncio.get_event_loop().time()

        try:
            while True:
                try:
                    # Wait for a message with timeout
                    message = await asyncio.wait_for(
                        websocket.receive_text(), timeout=HEARTBEAT_TIMEOUT
                    )

                    try:
                        data = json.loads(message)
                        if data.get("type") == "ping":
                            # Respond to heartbeat
                            await websocket.send_text(json.dumps({"type": "pong"}))
                            last_heartbeat = asyncio.get_event_loop().time()
                            continue
                    except json.JSONDecodeError:
                        # Not a JSON message, ignore
                        pass

                except TimeoutError:
                    # Check if we've exceeded the heartbeat timeout
                    if asyncio.get_event_loop().time() - last_heartbeat > HEARTBEAT_TIMEOUT:
                        logging.warning(f"Heartbeat timeout for peer {peer_id}")
                        break
                    continue
                except WebSocketDisconnect:
                    logging.info(f"Client {peer_id} disconnected")
                    break

        finally:
            # Clean up
            if peer_id in ws_map:
                del ws_map[peer_id]
            if not bot_task.done():
                bot_task.cancel()
                try:
                    await bot_task
                except asyncio.CancelledError:
                    pass
            logging.info(f"WebSocket connection closed for peer {peer_id}")

    except Exception as e:
        logging.error(f"Error in websocket endpoint: {str(e)}", exc_info=True)
        if peer_id in ws_map:
            del ws_map[peer_id]


"""
python -m src.cmd.webrtc_websocket.fastapi_ws_signaling_bot_serve -f config/bots/small_webrtc_fastapi_websocket_avatar_echo_bot.json
"""

if __name__ == "__main__":
    import uvicorn

    default_host = os.getenv("HOST", "0.0.0.0")
    default_port = int(os.getenv("FAST_API_PORT", "4321"))

    parser = argparse.ArgumentParser(description="Fastapi Websocket Bot Runner")
    parser.add_argument("--host", type=str, default=default_host, help="Host address")
    parser.add_argument("--port", type=int, default=default_port, help="Port number")
    parser.add_argument("--reload", action="store_true", help="Reload code on change")
    parser.add_argument("--ngrok", action="store_true", help="use ngrok proxy")
    parser.add_argument(
        "-f",
        type=str,
        default=os.path.join(CONFIG_DIR, "bots/dummy_bot.json"),
        help="Bot configuration json file",
    )

    config = parser.parse_args()

    if config.ngrok:
        ngrok_proxy(config.port)

    # Note: not event loop to new one
    # run_bot: AIFastapiWebsocketBot = asyncio.get_event_loop().run_until_complete(BotLoader.load_bot(config.f, bot_type="fastapi_ws_bot"))

    # use one event loop to run
    # run_bot: AIFastapiWebsocketBot = asyncio.run(
    #    BotLoader.load_bot(config.f, bot_type="fastapi_ws_bot"))

    # api docs: http://0.0.0.0:4321/docs
    uvicorn.run(
        app,
        host=config.host,
        port=config.port,
        reload=config.reload,
    )



# Copyright (c) 2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import onnxruntime
import torch
import numpy as np
import whisper
from typing import Callable
import torchaudio.compliance.kaldi as kaldi


class CosyVoiceFrontEnd:

    def __init__(
        self,
        feat_extractor: Callable,
        campplus_model: str,
        speech_tokenizer_model: str,
    ):
        self.feat_extractor = feat_extractor
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        option = onnxruntime.SessionOptions()
        option.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        )
        option.intra_op_num_threads = 1
        self.campplus_session = onnxruntime.InferenceSession(
            campplus_model, sess_options=option, providers=["CPUExecutionProvider"]
        )
        self.speech_tokenizer_session = onnxruntime.InferenceSession(
            speech_tokenizer_model,
            sess_options=option,
            providers=[
                (
                    "CUDAExecutionProvider"
                    if torch.cuda.is_available()
                    else "CPUExecutionProvider"
                )
            ],
        )

    def _extract_speech_token(self, speech):
        assert (
            speech.shape[1] / 16000 <= 30
        ), "do not support extract speech token for audio longer than 30s"
        feat = whisper.log_mel_spectrogram(speech, n_mels=128)
        speech_token = (
            self.speech_tokenizer_session.run(
                None,
                {
                    self.speech_tokenizer_session.get_inputs()[0]
                    .name: feat.detach()
                    .cpu()
                    .numpy(),
                    self.speech_tokenizer_session.get_inputs()[1].name: np.array(
                        [feat.shape[2]], dtype=np.int32
                    ),
                },
            )[0]
            .flatten()
            .tolist()
        )
        speech_token = torch.tensor([speech_token], dtype=torch.int32).to(self.device)
        speech_token_len = torch.tensor([speech_token.shape[1]], dtype=torch.int32).to(
            self.device
        )
        return speech_token, speech_token_len

    def _extract_spk_embedding(self, speech):
        feat = kaldi.fbank(speech, num_mel_bins=80, dither=0, sample_frequency=16000)
        feat = feat - feat.mean(dim=0, keepdim=True)
        embedding = (
            self.campplus_session.run(
                None,
                {
                    self.campplus_session.get_inputs()[0]
                    .name: feat.unsqueeze(dim=0)
                    .cpu()
                    .numpy()
                },
            )[0]
            .flatten()
            .tolist()
        )
        embedding = torch.tensor([embedding]).to(self.device)
        return embedding

    def _extract_speech_feat(self, speech):
        speech_feat = (
            self.feat_extractor(speech).squeeze(dim=0).transpose(0, 1).to(self.device)
        )
        speech_feat = speech_feat.unsqueeze(dim=0)
        speech_feat_len = torch.tensor([speech_feat.shape[1]], dtype=torch.int32).to(
            self.device
        )
        return speech_feat, speech_feat_len



from achatbot.modules.punctuation import PuncEnvInit
from achatbot.common.session import Session, SessionCtx

engine = PuncEnvInit.initEngine()
session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
# session.ctx.state["punc_cache"] = {}
session.ctx.state.update({"punc_cache": {}})
inputs = "你好|你叫什么名字|讲一个故事|境河流是养育沿岸|人民的生命之源长期以来为帮助下游地区防灾减灾中方技术人员|在上游地区极为恶劣的自然条件下克服巨大困难甚至冒着生命危险|向印方提供汛期水文资料处理紧急事件中方重视印方在跨境河流问题上的关切|愿意进一步完善双方联合工作机制|凡是|中方能做的我们|都会去做而且会做得更好我请印度朋友们放心中国在上游的|任何开发利用都会经过科学|规划和论证兼顾上下游的利益"
inputs = "你好|你叫什么名字|讲一个故事|"
vads = inputs.split("|")
rec_result_all = "outputs: "
for text in vads:
    if not text:
        continue
    session.ctx.state["text"] = text
    punc_text = engine.generate(session)
    print(session.ctx.state.get("punc_cache"), punc_text)
    rec_result_all += punc_text

print(rec_result_all)

"""
# pytorch
python -m src.modules.punctuation
PUNC_TAG=punc_ct_tranformer_offline python -m src.modules.punctuation

# onnx
PUNC_TAG=punc_ct_tranformer_onnx python -m src.modules.punctuation
PUNC_TAG=punc_ct_tranformer_onnx_offline python -m src.modules.punctuation

"""



from modelscope import AutoModel
import typer

app = typer.Typer()


@app.command()
def online():
    """
    https://modelscope.cn/models/iic/punc_ct-transformer_zh-cn-common-vad_realtime-vocab272727/summary
    """
    model = AutoModel(model="iic/punc_ct-transformer_zh-cn-common-vad_realtime-vocab272727")
    print(model.model)

    inputs = "跨境河流是养育沿岸|人民的生命之源长期以来为帮助下游地区防灾减灾中方技术人员|在上游地区极为恶劣的自然条件下克服巨大困难甚至冒着生命危险|向印方提供汛期水文资料处理紧急事件中方重视印方在跨境河流问题上的关切|愿意进一步完善双方联合工作机制|凡是|中方能做的我们|都会去做而且会做得更好我请印度朋友们放心中国在上游的|任何开发利用都会经过科学|规划和论证兼顾上下游的利益"
    vads = inputs.split("|")
    rec_result_all = "outputs: "
    cache = {}
    for vad in vads:
        rec_result = model.generate(input=vad, cache=cache, disable_pbar=True)
        print(rec_result)
        rec_result_all += rec_result[0]["text"]

    print(rec_result_all)


@app.command()
def offline():
    """
    - https://modelscope.cn/models/iic/punc_ct-transformer_zh-cn-common-vocab272727-pytorch/summary
    - https://modelscope.cn/models/iic/punc_ct-transformer_cn-en-common-vocab471067-large/summary
    """
    model = AutoModel(
        model="iic/punc_ct-transformer_zh-cn-common-vocab272727-pytorch",
        model_revision=None,
    )
    print(model.model)

    res = model.generate(input="那今天的会就到这里吧 happy new year 明年见", disable_pbar=True)
    print(res)


"""
python demo/modelscope/ct_tranfromer_punc.py offline
python demo/modelscope/ct_tranfromer_punc.py online
"""
if __name__ == "__main__":
    app()


"""
CTTransformerStreaming(
  (embed): Embedding(272727, 256)
  (decoder): Linear(in_features=256, out_features=6, bias=True)
  (encoder): SANMVadEncoder(
    (embed): SinusoidalPositionEncoder()
    (encoders0): MultiSequential(
      (0): EncoderLayerSANM(
        (self_attn): MultiHeadedAttentionSANMwithMask(
          (linear_out): Linear(in_features=256, out_features=256, bias=True)
          (linear_q_k_v): Linear(in_features=256, out_features=768, bias=True)
          (dropout): Dropout(p=0.0, inplace=False)
          (fsmn_block): Conv1d(256, 256, kernel_size=(11,), stride=(1,), groups=256, bias=False)
          (pad_fn): ConstantPad1d(padding=(10, 0), value=0.0)
        )
        (feed_forward): PositionwiseFeedForward(
          (w_1): Linear(in_features=256, out_features=1024, bias=True)
          (w_2): Linear(in_features=1024, out_features=256, bias=True)
          (dropout): Dropout(p=0.1, inplace=False)
          (activation): ReLU()
        )
        (norm1): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (norm2): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (dropout): Dropout(p=0.1, inplace=False)
      )
    )
    (encoders): MultiSequential(
      (0): EncoderLayerSANM(
        (self_attn): MultiHeadedAttentionSANMwithMask(
          (linear_out): Linear(in_features=256, out_features=256, bias=True)
          (linear_q_k_v): Linear(in_features=256, out_features=768, bias=True)
          (dropout): Dropout(p=0.0, inplace=False)
          (fsmn_block): Conv1d(256, 256, kernel_size=(11,), stride=(1,), groups=256, bias=False)
          (pad_fn): ConstantPad1d(padding=(10, 0), value=0.0)
        )
        (feed_forward): PositionwiseFeedForward(
          (w_1): Linear(in_features=256, out_features=1024, bias=True)
          (w_2): Linear(in_features=1024, out_features=256, bias=True)
          (dropout): Dropout(p=0.1, inplace=False)
          (activation): ReLU()
        )
        (norm1): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (norm2): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (dropout): Dropout(p=0.1, inplace=False)
      )
      (1): EncoderLayerSANM(
        (self_attn): MultiHeadedAttentionSANMwithMask(
          (linear_out): Linear(in_features=256, out_features=256, bias=True)
          (linear_q_k_v): Linear(in_features=256, out_features=768, bias=True)
          (dropout): Dropout(p=0.0, inplace=False)
          (fsmn_block): Conv1d(256, 256, kernel_size=(11,), stride=(1,), groups=256, bias=False)
          (pad_fn): ConstantPad1d(padding=(10, 0), value=0.0)
        )
        (feed_forward): PositionwiseFeedForward(
          (w_1): Linear(in_features=256, out_features=1024, bias=True)
          (w_2): Linear(in_features=1024, out_features=256, bias=True)
          (dropout): Dropout(p=0.1, inplace=False)
          (activation): ReLU()
        )
        (norm1): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (norm2): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
        (dropout): Dropout(p=0.1, inplace=False)
      )
    )
    (after_norm): LayerNorm((256,), eps=1e-12, elementwise_affine=True)
    (dropout): Dropout(p=0.1, inplace=False)
  )
)
"""



import logging
import asyncio
import time
from typing import Any

from apipeline.frames.base import Frame
from apipeline.frames.control_frames import StartFrame
from apipeline.processors.frame_processor import FrameDirection

from achatbot.processors.audio_input_processor import AudioVADInputProcessor
from achatbot.services.daily_client import DailyTransportClient
from achatbot.types.frames.data_frames import (
    InterimTranscriptionFrame,
    TranscriptionFrame,
    UserImageRawFrame,
)
from achatbot.types.frames.control_frames import UserImageRequestFrame
from achatbot.common.types import DailyParams
from achatbot.types.frames.data_frames import DailyTransportMessageFrame
from achatbot.modules.speech.vad_analyzer.daily_webrtc import DailyWebRTCVADAnalyzer


class DailyInputTransportProcessor(AudioVADInputProcessor):
    def __init__(self, client: DailyTransportClient, params: DailyParams, **kwargs):
        super().__init__(params, **kwargs)

        if params.vad_enabled and not params.vad_analyzer:
            # Default use DailyWebRTCVADAnalyzer
            self._vad_analyzer = DailyWebRTCVADAnalyzer(
                sample_rate=self._params.audio_in_sample_rate,
                num_channels=self._params.audio_in_channels,
            )

        self._client = client
        self._video_renderers = {}

    async def start(self, frame: StartFrame):
        # Parent start.
        await super().start(frame)
        # Join the room.
        await self._client.join()
        # Create audio task. It reads audio frames from Daily and push them
        # internally for VAD processing.
        if self._params.audio_in_enabled or self._params.vad_enabled:
            self._audio_in_task = self.get_event_loop().create_task(self._audio_in_task_handler())

    async def stop(self):
        # Stop audio thread.
        if self._params.audio_in_enabled or self._params.vad_enabled:
            if self._audio_in_task:
                self._audio_in_task.cancel()
                await self._audio_in_task
        # Leave the room.
        await self._client.leave()
        await self.cleanup()
        # Parent stop.
        await super().stop()

    async def cleanup(self):
        await super().cleanup()
        await self._client.cleanup()

    #
    # FrameProcessor
    #

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, UserImageRequestFrame):
            self.request_participant_image(frame.user_id)

    #
    # Frames
    #

    async def push_transcription_frame(self, frame: TranscriptionFrame | InterimTranscriptionFrame):
        await self.queue_frame(frame)

    async def push_app_message(self, message: Any, sender: str):
        frame = DailyTransportMessageFrame(message=message, participant_id=sender)
        await self.queue_frame(frame)

    #
    # Audio in
    #

    async def _audio_in_task_handler(self):
        while True:
            try:
                frame = await self._client.read_next_audio_frame()
                if frame:
                    await self.push_audio_frame(frame)
            except asyncio.CancelledError:
                break

    #
    # Camera in
    #

    def capture_participant_video(
        self,
        participant_id: str,
        framerate: int = 30,
        video_source: str = "camera",
        color_format: str = "RGB",
    ):
        self._video_renderers[participant_id] = {
            "framerate": framerate,
            "timestamp": 0,
            "render_next_frame": False,
        }

        self._client.capture_participant_video(
            participant_id, self._on_participant_video_frame, framerate, video_source, color_format
        )

    def request_participant_image(self, participant_id: str):
        if participant_id in self._video_renderers:
            self._video_renderers[participant_id]["render_next_frame"] = True

    async def _on_participant_video_frame(self, participant_id: str, buffer, size, color_format):
        """
        control render image frame rate or request participant image to render
        """
        render_frame = False

        curr_time = time.time()
        prev_time = self._video_renderers[participant_id]["timestamp"] or curr_time
        framerate = self._video_renderers[participant_id]["framerate"]

        if framerate > 0:
            next_time = prev_time + 1 / framerate
            render_frame = (curr_time - next_time) < 0.1
        elif self._video_renderers[participant_id]["render_next_frame"]:
            # e.g.: push UserImageRequestFrame to render a UserImageRawFrame
            self._video_renderers[participant_id]["render_next_frame"] = False
            render_frame = True

        if render_frame:
            frame = UserImageRawFrame(
                image=buffer,
                size=size,
                format="JPEG",  # from frame bytes, no save format, need add a save format e.g.: JPEG,PNG,
                mode=color_format,  # default: RGB
                user_id=participant_id,
            )
            await self.queue_frame(frame)

        self._video_renderers[participant_id]["timestamp"] = curr_time



import os
import logging
import asyncio
import time

from fastapi import WebSocket
from src.cmd.bots.base import AIBot
from src.common.interface import IBot
from . import register_ai_room_bots, register_ai_fastapi_ws_bots


@register_ai_room_bots.register
@register_ai_fastapi_ws_bots.register
class DummyBot(AIBot):
    def __init__(self, **args) -> None:
        self._websocket = args.pop("websocket", None)
        super().__init__(**args)
        self.init_bot_config()

    def set_fastapi_websocket(self, websocket: WebSocket):
        self._websocket = websocket

    async def arun(self):
        logging.info(f"Starting bot,env: {os.environ}")
        logging.info(f"dummy bot run with args: {self.args}")
        # time.sleep(10)
        await asyncio.sleep(10)
        logging.info("dummy bot over")
